# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Preferences plugin public facing API
"""

# Standard library imports
import ast
import os.path as osp

# Third party imports
from qtpy import API
from qtpy.compat import (getexistingdirectory, getopenfilename, from_qvariant,
                         to_qvariant)
from qtpy.QtCore import Qt, QRegularExpression, QSize, Signal, Slot
from qtpy.QtGui import QColor, QRegularExpressionValidator, QTextOption
from qtpy.QtWidgets import (
    QAction,
    QButtonGroup,
    QCheckBox,
    QDoubleSpinBox,
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QMessageBox,
    QPlainTextEdit,
    QPushButton,
    QRadioButton,
    QSpinBox,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.widgets.comboboxes import SpyderComboBox, SpyderFontComboBox
from spyder.config.base import _
from spyder.config.manager import CONF
from spyder.config.user import NoDefault
from spyder.py3compat import to_text_string
from spyder.utils.icon_manager import ima
from spyder.utils.misc import getcwd_or_home
from spyder.utils.stylesheet import AppStyle, MAC, WIN
from spyder.widgets.colors import ColorLayout
from spyder.widgets.helperwidgets import TipWidget, ValidationLineEdit
from spyder.widgets.comboboxes import FileComboBox
from spyder.widgets.sidebardialog import SidebarPage


class BaseConfigTab(QWidget):
    """Stub class to declare a config tab."""
    pass


class ConfigAccessMixin:
    """Mixin to access config options in SpyderConfigPages."""
    CONF_SECTION = None

    def set_option(
        self,
        option,
        value,
        section=None,
        recursive_notification=False,
        secure=False,
    ):
        section = self.CONF_SECTION if section is None else section
        CONF.set(
            section,
            option,
            value,
            recursive_notification=recursive_notification,
            secure=secure,
        )

    def get_option(
        self, option, default=NoDefault, section=None, secure=False
    ):
        section = self.CONF_SECTION if section is None else section
        return CONF.get(section, option, default=default, secure=secure)

    def remove_option(self, option, section=None, secure=False):
        section = self.CONF_SECTION if section is None else section
        CONF.remove_option(section, option, secure=secure)


class SpyderConfigPage(SidebarPage, ConfigAccessMixin):
    """
    Page that can display graphical elements connected to our config system.
    """

    # Signals
    apply_button_enabled = Signal(bool)

    # Constants
    CONF_SECTION = None
    LOAD_FROM_CONFIG = True

    def __init__(self, parent):
        SidebarPage.__init__(self, parent)

        # Callback to call before saving settings to disk
        self.pre_apply_callback = None

        # Callback to call after saving settings to disk
        self.apply_callback = lambda: self._apply_settings_tabs(
            self.changed_options
        )

        self.checkboxes = {}
        self.radiobuttons = {}
        self.lineedits = {}
        self.textedits = {}
        self.validate_data = {}
        self.spinboxes = {}
        self.comboboxes = {}
        self.fontboxes = {}
        self.coloredits = {}
        self.scedits = {}
        self.cross_section_options = {}

        self.changed_options = set()
        self.restart_options = dict()  # Dict to store name and localized text
        self.default_button_group = None
        self.tabs = None
        self.is_modified = False

        if getattr(parent, "main", None):
            self.main = parent.main
        else:
            self.main = None

    def initialize(self):
        """Initialize configuration page."""
        self.setup_page()
        if self.LOAD_FROM_CONFIG:
            self.load_from_conf()

    def _apply_settings_tabs(self, options):
        if self.tabs is not None:
            for i in range(self.tabs.count()):
                tab = self.tabs.widget(i)
                layout = tab.layout()
                for i in range(layout.count()):
                    widget = layout.itemAt(i).widget()
                    if hasattr(widget, 'apply_settings'):
                        if issubclass(type(widget), BaseConfigTab):
                            options |= widget.apply_settings()
        self.apply_settings(options)

    def apply_settings(self, options):
        raise NotImplementedError

    def apply_changes(self):
        """Apply changes callback"""
        if self.is_modified:
            if self.pre_apply_callback is not None:
                self.pre_apply_callback()

            self.save_to_conf()

            if self.apply_callback is not None:
                self.apply_callback()

            # Since the language cannot be retrieved by CONF and the language
            # is needed before loading CONF, this is an extra method needed to
            # ensure that when changes are applied, they are copied to a
            # specific file storing the language value. This only applies to
            # the main section config.
            if self.CONF_SECTION == 'main':
                self._save_lang()

            for restart_option in self.restart_options:
                if restart_option in self.changed_options:
                    self.prompt_restart_required()
                    break  # Ensure a single popup is displayed
            self.set_modified(False)

    def check_settings(self):
        """This method is called to check settings after configuration
        dialog has been shown"""
        pass

    def set_modified(self, state):
        self.is_modified = state
        self.apply_button_enabled.emit(state)
        if not state:
            self.changed_options = set()

    def is_valid(self):
        """Return True if all widget contents are valid"""
        status = True
        for lineedit in self.lineedits:
            if lineedit in self.validate_data and lineedit.isEnabled():
                validator, invalid_msg = self.validate_data[lineedit]
                text = to_text_string(lineedit.text())
                if not validator(text):
                    QMessageBox.critical(self, self.get_name(),
                                         f"{invalid_msg}:<br><b>{text}</b>",
                                         QMessageBox.Ok)
                    return False

        if self.tabs is not None and status:
            for i in range(self.tabs.count()):
                tab = self.tabs.widget(i)
                layout = tab.layout()
                for i in range(layout.count()):
                    widget = layout.itemAt(i).widget()
                    if issubclass(type(widget), BaseConfigTab):
                        status &= widget.is_valid()
                        if not status:
                            return status
        return status

    def reset_widget_dicts(self):
        """Reset the dicts of widgets tracked in the page."""
        self.checkboxes = {}
        self.radiobuttons = {}
        self.lineedits = {}
        self.textedits = {}
        self.validate_data = {}
        self.spinboxes = {}
        self.comboboxes = {}
        self.fontboxes = {}
        self.coloredits = {}
        self.scedits = {}
        self.cross_section_options = {}

    def load_from_conf(self):
        """Load settings from configuration file."""
        for checkbox, (sec, option, default) in list(self.checkboxes.items()):
            checkbox.setChecked(self.get_option(option, default, section=sec))
            checkbox.clicked[bool].connect(lambda _, opt=option, sect=sec:
                                           self.has_been_modified(sect, opt))
            if checkbox.restart_required:
                if sec is None:
                    self.restart_options[option] = checkbox.text()
                else:
                    self.restart_options[(sec, option)] = checkbox.text()

        for radiobutton, (sec, option, default) in list(
                self.radiobuttons.items()):
            radiobutton.setChecked(self.get_option(option, default,
                                                   section=sec))
            radiobutton.toggled.connect(lambda _foo, opt=option, sect=sec:
                                        self.has_been_modified(sect, opt))
            if radiobutton.restart_required:
                if sec is None:
                    self.restart_options[option] = radiobutton.label_text
                else:
                    self.restart_options[(sec, option)] = radiobutton.label_text

        for lineedit, (sec, option, default) in list(self.lineedits.items()):
            data = self.get_option(
                option,
                default,
                section=sec,
                secure=True
                if (hasattr(lineedit, "password") and lineedit.password)
                else False,
            )

            if getattr(lineedit, 'content_type', None) == list:
                data = ', '.join(data)
            else:
                # Make option value a string to prevent errors when using it
                # as widget text.
                # See spyder-ide/spyder#18929
                data = str(data)
            lineedit.setText(data)
            lineedit.textChanged.connect(lambda _, opt=option, sect=sec:
                                         self.has_been_modified(sect, opt))
            if lineedit.restart_required:
                if sec is None:
                    self.restart_options[option] = lineedit.label_text
                else:
                    self.restart_options[(sec, option)] = lineedit.label_text

        for textedit, (sec, option, default) in list(self.textedits.items()):
            data = self.get_option(option, default, section=sec)
            if getattr(textedit, 'content_type', None) == list:
                data = ', '.join(data)
            elif getattr(textedit, 'content_type', None) == dict:
                data = to_text_string(data)
            textedit.setPlainText(data)
            textedit.textChanged.connect(lambda opt=option, sect=sec:
                                         self.has_been_modified(sect, opt))
            if textedit.restart_required:
                if sec is None:
                    self.restart_options[option] = textedit.label_text
                else:
                    self.restart_options[(sec, option)] = textedit.label_text

        for spinbox, (sec, option, default) in list(self.spinboxes.items()):
            spinbox.setValue(self.get_option(option, default, section=sec))
            spinbox.valueChanged.connect(lambda _foo, opt=option, sect=sec:
                                         self.has_been_modified(sect, opt))

        for combobox, (sec, option, default) in list(self.comboboxes.items()):
            value = self.get_option(option, default, section=sec)
            for index in range(combobox.count()):
                data = from_qvariant(combobox.itemData(index), to_text_string)
                # For PyQt API v2, it is necessary to convert `data` to
                # unicode in case the original type was not a string, like an
                # integer for example (see qtpy.compat.from_qvariant):
                if to_text_string(data) == to_text_string(value):
                    break
            else:
                if combobox.count() == 0:
                    index = None
            if index:
                combobox.setCurrentIndex(index)
            combobox.currentIndexChanged.connect(
                lambda _foo, opt=option, sect=sec:
                    self.has_been_modified(sect, opt))
            if combobox.restart_required:
                if sec is None:
                    self.restart_options[option] = combobox.label_text
                else:
                    self.restart_options[(sec, option)] = combobox.label_text

        for (fontbox, sizebox), option in list(self.fontboxes.items()):
            font = self.get_font(option)
            fontbox.setCurrentFont(font)
            sizebox.setValue(font.pointSize())

            fontbox.currentIndexChanged.connect(
                lambda _foo, opt=option: self.has_been_modified(None, opt))
            sizebox.valueChanged.connect(
                lambda _foo, opt=option: self.has_been_modified(None, opt))

            if fontbox.restart_required:
                self.restart_options[option] = fontbox.label_text

            if sizebox.restart_required:
                self.restart_options[option] = sizebox.label_text

        for clayout, (sec, option, default) in list(self.coloredits.items()):
            edit = clayout.lineedit
            btn = clayout.colorbtn
            edit.setText(self.get_option(option, default, section=sec))
            # QAbstractButton works differently for PySide and PyQt
            if not API == 'pyside':
                btn.clicked.connect(lambda _foo, opt=option, sect=sec:
                                    self.has_been_modified(sect, opt))
            else:
                btn.clicked.connect(lambda opt=option, sect=sec:
                                    self.has_been_modified(sect, opt))
            edit.textChanged.connect(lambda _foo, opt=option, sect=sec:
                                     self.has_been_modified(sect, opt))

        for (clayout, cb_bold, cb_italic
             ), (sec, option, default) in list(self.scedits.items()):
            edit = clayout.lineedit
            btn = clayout.colorbtn
            options = self.get_option(option, default, section=sec)
            if options:
                color, bold, italic = options
                edit.setText(color)
                cb_bold.setChecked(bold)
                cb_italic.setChecked(italic)

            edit.textChanged.connect(lambda _foo, opt=option, sect=sec:
                                     self.has_been_modified(sect, opt))
            btn.clicked[bool].connect(lambda _foo, opt=option, sect=sec:
                                      self.has_been_modified(sect, opt))
            cb_bold.clicked[bool].connect(lambda _foo, opt=option, sect=sec:
                                          self.has_been_modified(sect, opt))
            cb_italic.clicked[bool].connect(lambda _foo, opt=option, sect=sec:
                                            self.has_been_modified(sect, opt))

    def save_to_conf(self):
        """Save settings to configuration file"""
        for checkbox, (sec, option, _default) in list(
                self.checkboxes.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                value = checkbox.isChecked()
                self.set_option(option, value, section=sec,
                                recursive_notification=False)

        for radiobutton, (sec, option, _default) in list(
                self.radiobuttons.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                self.set_option(option, radiobutton.isChecked(), section=sec,
                                recursive_notification=False)

        for lineedit, (sec, option, _default) in list(self.lineedits.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                data = lineedit.text()
                content_type = getattr(lineedit, 'content_type', None)
                if content_type == list:
                    data = [item.strip() for item in data.split(',')]
                else:
                    data = to_text_string(data)

                self.set_option(
                    option,
                    data,
                    section=sec,
                    recursive_notification=False,
                    secure=True
                    if (hasattr(lineedit, "password") and lineedit.password)
                    else False,
                )

        for textedit, (sec, option, _default) in list(self.textedits.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                data = textedit.toPlainText()
                content_type = getattr(textedit, 'content_type', None)
                if content_type == dict:
                    if data:
                        data = ast.literal_eval(data)
                    else:
                        data = textedit.content_type()
                elif content_type in (tuple, list):
                    data = [item.strip() for item in data.split(',')]
                else:
                    data = to_text_string(data)
                self.set_option(option, data, section=sec,
                                recursive_notification=False)

        for spinbox, (sec, option, _default) in list(self.spinboxes.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                self.set_option(option, spinbox.value(), section=sec,
                                recursive_notification=False)

        for combobox, (sec, option, _default) in list(self.comboboxes.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                data = combobox.itemData(combobox.currentIndex())
                self.set_option(option, from_qvariant(data, to_text_string),
                                section=sec, recursive_notification=False)

        for (fontbox, sizebox), option in list(self.fontboxes.items()):
            if option in self.changed_options or not self.LOAD_FROM_CONFIG:
                font = fontbox.currentFont()
                font.setPointSize(sizebox.value())
                self.set_font(font, option)

        for clayout, (sec, option, _default) in list(self.coloredits.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                self.set_option(option,
                                to_text_string(clayout.lineedit.text()),
                                section=sec, recursive_notification=False)

        for (clayout, cb_bold, cb_italic), (sec, option, _default) in list(
                self.scedits.items()):
            if (
                option in self.changed_options
                or (sec, option) in self.changed_options
                or not self.LOAD_FROM_CONFIG
            ):
                color = to_text_string(clayout.lineedit.text())
                bold = cb_bold.isChecked()
                italic = cb_italic.isChecked()
                self.set_option(option, (color, bold, italic), section=sec,
                                recursive_notification=False)

    @Slot(str)
    def has_been_modified(self, section, option):
        self.set_modified(True)
        if section is None:
            self.changed_options.add(option)
        else:
            self.changed_options.add((section, option))

    def add_help_info_label(self, layout, tip_text):
        help_label = TipWidget(
            tip_text=tip_text,
            icon=ima.icon('question_tip'),
            hover_icon=ima.icon('question_tip_hover')
        )

        layout.addWidget(help_label)
        layout.addStretch(100)

        return layout, help_label

    def create_checkbox(self, text, option, default=NoDefault,
                        tip=None, msg_warning=None, msg_info=None,
                        msg_if_enabled=False, section=None, restart=False):
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        checkbox = QCheckBox(text)
        layout.addWidget(checkbox)

        self.checkboxes[checkbox] = (section, option, default)
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        if msg_warning is not None or msg_info is not None:
            def show_message(is_checked=False):
                if is_checked or not msg_if_enabled:
                    if msg_warning is not None:
                        QMessageBox.warning(self, self.get_name(),
                                            msg_warning, QMessageBox.Ok)
                    if msg_info is not None:
                        QMessageBox.information(self, self.get_name(),
                                                msg_info, QMessageBox.Ok)
            checkbox.clicked.connect(show_message)
        checkbox.restart_required = restart

        widget = QWidget(self)
        widget.checkbox = checkbox
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            widget.help_label = help_label
        widget.setLayout(layout)
        return widget

    def create_radiobutton(self, text, option, default=NoDefault,
                           tip=None, msg_warning=None, msg_info=None,
                           msg_if_enabled=False, button_group=None,
                           restart=False, section=None):
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        radiobutton = QRadioButton(text)
        layout.addWidget(radiobutton)

        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        if button_group is None:
            if self.default_button_group is None:
                self.default_button_group = QButtonGroup(self)
            button_group = self.default_button_group
        button_group.addButton(radiobutton)
        self.radiobuttons[radiobutton] = (section, option, default)
        if msg_warning is not None or msg_info is not None:
            def show_message(is_checked):
                if is_checked or not msg_if_enabled:
                    if msg_warning is not None:
                        QMessageBox.warning(self, self.get_name(),
                                            msg_warning, QMessageBox.Ok)
                    if msg_info is not None:
                        QMessageBox.information(self, self.get_name(),
                                                msg_info, QMessageBox.Ok)
            radiobutton.toggled.connect(show_message)
        radiobutton.restart_required = restart
        radiobutton.label_text = text

        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            radiobutton.help_label = help_label
        widget = QWidget(self)
        widget.radiobutton = radiobutton
        widget.setLayout(layout)
        return widget

    def create_lineedit(self, text, option, default=NoDefault,
                        tip=None, alignment=Qt.Vertical, regex=None,
                        restart=False, word_wrap=True, placeholder=None,
                        content_type=None, section=None, status_icon=None,
                        password=False, validate_callback=None,
                        validate_reason=None):
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section

        label = QLabel(text)
        label.setWordWrap(word_wrap)

        if validate_callback:
            if not validate_reason:
                raise RuntimeError(
                    "You need to provide a validate_reason if you want to use "
                    "a validate_callback"
                )

            edit = ValidationLineEdit(
                validate_callback=validate_callback,
                validate_reason=validate_reason,
            )
            status_action = edit.error_action
        else:
            edit = QLineEdit()
        edit.content_type = content_type
        if password:
            edit.setEchoMode(QLineEdit.Password)

        if status_icon is not None:
            status_action = QAction(self)
            edit.addAction(status_action, QLineEdit.TrailingPosition)
            status_action.setIcon(status_icon)
            status_action.setVisible(False)

        if alignment == Qt.Vertical:
            layout = QVBoxLayout()

            # This is necessary to correctly align `label` and `edit` to the
            # left when they are displayed vertically.
            edit.setStyleSheet("margin-left: 5px")

            if tip is not None:
                label_layout = QHBoxLayout()
                label_layout.setSpacing(0)
                label_layout.addWidget(label)
                label_layout, help_label = self.add_help_info_label(
                    label_layout, tip
                )
                layout.addLayout(label_layout)
            else:
                layout.addWidget(label)

            layout.addWidget(edit)
        else:
            layout = QHBoxLayout()
            layout.addWidget(label)
            layout.addWidget(edit)
            if tip is not None:
                layout, help_label = self.add_help_info_label(layout, tip)

        layout.setContentsMargins(0, 0, 0, 0)

        if regex:
            edit.setValidator(
                QRegularExpressionValidator(QRegularExpression(regex))
            )

        if placeholder:
            edit.setPlaceholderText(placeholder)

        self.lineedits[edit] = (section, option, default)

        widget = QWidget(self)
        widget.label = label
        widget.textbox = edit
        if tip is not None:
            widget.help_label = help_label
        if status_icon is not None or validate_callback is not None:
            widget.status_action = status_action

        widget.setLayout(layout)
        edit.restart_required = restart
        edit.label_text = text
        edit.password = password

        return widget

    def create_textedit(self, text, option, default=NoDefault,
                        tip=None, restart=False, content_type=None,
                        section=None):
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        label = QLabel(text)
        label.setWordWrap(True)
        edit = QPlainTextEdit()
        edit.content_type = content_type
        edit.setWordWrapMode(QTextOption.WordWrap)
        layout = QVBoxLayout()
        layout.addWidget(label)
        layout.addWidget(edit)
        layout.setContentsMargins(0, 0, 0, 0)
        self.textedits[edit] = (section, option, default)

        widget = QWidget(self)
        widget.label = label
        widget.textbox = edit
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            widget.help_label = help_label
        widget.setLayout(layout)
        edit.restart_required = restart
        edit.label_text = text
        return widget

    def create_browsedir(self, text, option, default=NoDefault, section=None,
                         tip=None, alignment=Qt.Horizontal, status_icon=None):
        widget = self.create_lineedit(
            text,
            option,
            default,
            section=section,
            alignment=alignment,
            # We need the tip to be added by the lineedit if the alignment is
            # vertical. If not, it'll be added below when setting the layout.
            tip=tip if (tip and alignment == Qt.Vertical) else None,
            status_icon=status_icon,
        )

        for edit in self.lineedits:
            if widget.isAncestorOf(edit):
                break

        msg = _("Invalid directory path")
        self.validate_data[edit] = (osp.isdir, msg)

        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select directory"))
        browse_btn.clicked.connect(lambda: self.select_directory(edit))
        browse_btn.setIconSize(
            QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        if alignment == Qt.Vertical:
            # This is necessary to position browse_btn vertically centered with
            # respect to the lineedit.
            browse_btn.setStyleSheet("margin-top: 28px")

            layout = QGridLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget, 0, 0)
            layout.addWidget(browse_btn, 0, 1)
        else:
            # This is necessary to position browse_btn vertically centered with
            # respect to the lineedit.
            browse_btn.setStyleSheet("margin-top: 2px")

            layout = QHBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget)
            layout.addWidget(browse_btn)
            if tip is not None:
                layout, help_label = self.add_help_info_label(layout, tip)

        browsedir = QWidget(self)
        browsedir.textbox = widget.textbox
        if status_icon:
            browsedir.status_action = widget.status_action

        browsedir.setLayout(layout)
        return browsedir

    def select_directory(self, edit):
        """Select directory"""
        basedir = to_text_string(edit.text())
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        title = _("Select directory")
        directory = getexistingdirectory(self, title, basedir)
        if directory:
            edit.setText(directory)

    def create_browsefile(self, text, option, default=NoDefault, section=None,
                          tip=None, filters=None, alignment=Qt.Horizontal,
                          status_icon=None):
        widget = self.create_lineedit(
            text,
            option,
            default,
            section=section,
            alignment=alignment,
            # We need the tip to be added by the lineedit if the alignment is
            # vertical. If not, it'll be added below when setting the layout.
            tip=tip if (tip and alignment == Qt.Vertical) else None,
            status_icon=status_icon,
        )

        for edit in self.lineedits:
            if widget.isAncestorOf(edit):
                break

        msg = _('Invalid file path')
        self.validate_data[edit] = (osp.isfile, msg)

        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select file"))
        browse_btn.clicked.connect(lambda: self.select_file(edit, filters))
        browse_btn.setIconSize(
           QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        if alignment == Qt.Vertical:
            # This is necessary to position browse_btn vertically centered with
            # respect to the lineedit.
            browse_btn.setStyleSheet("margin-top: 28px")

            layout = QGridLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget, 0, 0)
            layout.addWidget(browse_btn, 0, 1)
        else:
            # This is necessary to position browse_btn vertically centered with
            # respect to the lineedit.
            browse_btn.setStyleSheet("margin-top: 2px")

            layout = QHBoxLayout()
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(widget)
            layout.addWidget(browse_btn)
            if tip is not None:
                layout, help_label = self.add_help_info_label(layout, tip)

        browsefile = QWidget(self)
        browsefile.textbox = widget.textbox
        if status_icon:
            browsefile.status_action = widget.status_action

        browsefile.setLayout(layout)
        return browsefile

    def select_file(self, edit, filters=None, **kwargs):
        """Select File"""
        basedir = osp.dirname(to_text_string(edit.text()))
        if not osp.isdir(basedir):
            basedir = getcwd_or_home()
        if filters is None:
            filters = _("All files (*)")
        title = _("Select file")
        filename, _selfilter = getopenfilename(self, title, basedir, filters,
                                               **kwargs)
        if filename:
            edit.setText(filename)

    def create_spinbox(self, prefix, suffix, option, default=NoDefault,
                       min_=None, max_=None, step=None, tip=None,
                       section=None):
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        widget = QWidget(self)
        if prefix:
            plabel = QLabel(prefix)
            widget.plabel = plabel
        else:
            plabel = None
        if suffix:
            slabel = QLabel(suffix)
            widget.slabel = slabel
        else:
            slabel = None
        if step is not None:
            if type(step) is int:
                spinbox = QSpinBox()
            else:
                spinbox = QDoubleSpinBox()
                spinbox.setDecimals(1)
            spinbox.setSingleStep(step)
        else:
            spinbox = QSpinBox()
        if min_ is not None:
            spinbox.setMinimum(min_)
        if max_ is not None:
            spinbox.setMaximum(max_)
        self.spinboxes[spinbox] = (section, option, default)
        layout = QHBoxLayout()
        for subwidget in (plabel, spinbox, slabel):
            if subwidget is not None:
                layout.addWidget(subwidget)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        widget.spinbox = spinbox
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            widget.help_label = help_label
        widget.setLayout(layout)
        return widget

    def create_coloredit(self, text, option, default=NoDefault, tip=None,
                         without_layout=False, section=None):
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        label = QLabel(text)
        clayout = ColorLayout(QColor(Qt.black), self)
        clayout.lineedit.setMaximumWidth(80)
        self.coloredits[clayout] = (section, option, default)
        if without_layout:
            return label, clayout
        layout = QHBoxLayout()
        layout.addWidget(label)
        layout.addLayout(clayout)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)

        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def create_scedit(self, text, option, default=NoDefault, tip=None,
                      without_layout=False, section=None):
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        label = QLabel(text)
        clayout = ColorLayout(QColor(Qt.black), self)
        clayout.lineedit.setMaximumWidth(80)
        cb_bold = QCheckBox()
        cb_bold.setIcon(ima.icon('bold'))
        cb_bold.setToolTip(_("Bold"))
        cb_italic = QCheckBox()
        cb_italic.setIcon(ima.icon('italic'))
        cb_italic.setToolTip(_("Italic"))
        self.scedits[(clayout, cb_bold, cb_italic)] = (section, option,
                                                       default)
        if without_layout:
            return label, clayout, cb_bold, cb_italic
        layout = QHBoxLayout()
        layout.addWidget(label)
        layout.addLayout(clayout)
        layout.addSpacing(10)
        layout.addWidget(cb_bold)
        layout.addWidget(cb_italic)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
        widget = QWidget(self)
        widget.setLayout(layout)
        return widget

    def create_combobox(self, text, choices, option, default=NoDefault,
                        tip=None, restart=False, section=None,
                        items_elide_mode=None):
        """choices: couples (name, key)"""
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        label = QLabel(text)
        combobox = SpyderComboBox(items_elide_mode=items_elide_mode)
        for name, key in choices:
            if not (name is None and key is None):
                combobox.addItem(name, to_qvariant(key))
        # Insert separators
        count = 0
        for index, item in enumerate(choices):
            name, key = item
            if name is None and key is None:
                combobox.insertSeparator(index + count)
                count += 1
        self.comboboxes[combobox] = (section, option, default)
        layout = QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(combobox)
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        widget = QWidget(self)
        widget.label = label
        widget.combobox = combobox
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            widget.help_label = help_label
        widget.setLayout(layout)
        combobox.restart_required = restart
        combobox.label_text = text
        return widget

    def create_file_combobox(self, text, choices, option, default=NoDefault,
                             tip=None, restart=False, filters=None,
                             adjust_to_contents=False,
                             default_line_edit=False, section=None,
                             validate_callback=None):
        """choices: couples (name, key)"""
        if section is not None and section != self.CONF_SECTION:
            self.cross_section_options[option] = section
        combobox = FileComboBox(self, adjust_to_contents=adjust_to_contents,
                                default_line_edit=default_line_edit)
        combobox.restart_required = restart
        combobox.label_text = text
        edit = combobox.lineEdit()
        edit.label_text = text
        edit.restart_required = restart
        self.lineedits[edit] = (section, option, default)
        combobox.addItems(choices)
        combobox.choices = choices

        msg = _('Invalid file path')
        self.validate_data[edit] = (
            validate_callback if validate_callback else osp.isfile,
            msg
        )

        browse_btn = QPushButton(ima.icon('DirOpenIcon'), '', self)
        browse_btn.setToolTip(_("Select file"))
        options = QFileDialog.DontResolveSymlinks
        browse_btn.clicked.connect(
            lambda: self.select_file(edit, filters, options=options)
        )
        browse_btn.setIconSize(
           QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
        )

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(combobox)
        layout.addWidget(browse_btn)
        layout.addStretch()

        widget = QWidget(self)
        widget.combobox = combobox
        widget.browse_btn = browse_btn
        if tip is not None:
            layout, help_label = self.add_help_info_label(layout, tip)
            widget.help_label = help_label
        widget.setLayout(layout)

        return widget

    def create_fontgroup(self, option=None, text=None, title=None,
                         tip=None, fontfilters=None, without_group=False,
                         restart=False):
        """Option=None -> setting plugin font"""

        if title:
            fontlabel = QLabel(title)
        else:
            fontlabel = QLabel(_("Font"))

        fontbox = SpyderFontComboBox()
        fontbox.restart_required = restart
        fontbox.label_text = _("{} font").format(title)

        if fontfilters is not None:
            fontbox.setFontFilters(fontfilters)

        sizebox = QSpinBox()
        sizebox.setRange(7, 100)
        sizebox.restart_required = restart
        sizebox.label_text = _("{} font size").format(title)

        self.fontboxes[(fontbox, sizebox)] = option

        layout = QHBoxLayout()
        for subwidget in (fontlabel, fontbox, sizebox):
            layout.addWidget(subwidget)
        layout.addStretch(1)

        if not without_group:
            if text is None:
                text = _("Font style")

            group = QGroupBox(text)
            group.setLayout(layout)

            if tip is not None:
                layout, help_label = self.add_help_info_label(layout, tip)

            return group
        else:
            widget = QWidget(self)
            widget.fontlabel = fontlabel
            widget.fontbox = fontbox
            widget.sizebox = sizebox
            widget.setLayout(layout)

            return widget

    def create_button(
        self,
        callback,
        text=None,
        icon=None,
        tooltip=None,
        set_modified_on_click=False,
    ):
        if icon is not None:
            btn = QPushButton(icon, "", parent=self)
            btn.setIconSize(
                QSize(AppStyle.ConfigPageIconSize, AppStyle.ConfigPageIconSize)
            )
        else:
            btn = QPushButton(text, parent=self)

        btn.clicked.connect(callback)
        if tooltip is not None:
            btn.setToolTip(tooltip)

        if set_modified_on_click:
            btn.clicked.connect(
                lambda checked=False, opt="": self.has_been_modified(
                    self.CONF_SECTION, opt
                )
            )

        return btn

    def create_tab(self, name, widgets):
        """
        Create a tab widget page.

        Parameters
        ----------
        name: str
            Name of the tab
        widgets: list or QWidget
            List of widgets to add to the tab. This can be also a single
            widget.

        Notes
        -----
        * Widgets are added in a vertical layout.
        """
        if self.tabs is None:
            self.tabs = QTabWidget(self)
            self.tabs.setUsesScrollButtons(True)
            self.tabs.setElideMode(Qt.ElideNone)

            vlayout = QVBoxLayout()
            vlayout.addWidget(self.tabs)
            self.setLayout(vlayout)

        if not isinstance(widgets, list):
            widgets = [widgets]

        tab = QWidget(self)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)

        # This is necessary to make Qt respect the declared vertical spacing
        # for widgets. In other words, it prevents text to be cropped when the
        # total height of the page is too large.
        layout.setSizeConstraint(QLayout.SetFixedSize)

        for w in widgets:
            # We need to set a min width so that pages are not shown too thin
            # due to setting the layout size constraint above.
            w.setMinimumWidth(
                self.MAX_WIDTH - (60 if MAC else (80 if WIN else 70))
            )
            layout.addWidget(w)

        layout.addStretch(1)
        tab.setLayout(layout)

        self.tabs.addTab(tab, name)

    def prompt_restart_required(self):
        """Prompt the user with a request to restart."""
        message = _(
            "One or more of the settings you changed requires a restart to be "
            "applied.<br><br>"
            "Do you wish to restart now?"
        )

        answer = QMessageBox.information(
            self,
            _("Information"),
            message,
            QMessageBox.Yes | QMessageBox.No
        )

        if answer == QMessageBox.Yes:
            self.restart()

    def restart(self):
        """Restart Spyder."""
        self.main.restart(close_immediately=True)

    def _add_tab(self, Widget):
        widget = Widget(self)

        if self.tabs is None:
            # In case a preference page does not have any tabs, we need to
            # add a tab with the widgets that already exist and then add the
            # new tab.
            layout = self.layout()
            main_widget = QWidget(self)
            main_widget.setLayout(layout)

            self.create_tab(_('General'), main_widget)
            self.create_tab(Widget.TITLE, widget)
        else:
            self.create_tab(Widget.TITLE, widget)

        self.load_from_conf()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Contains the bookmarsks utilities.
"""
# Standard imports
import os.path as osp


def _load_all_bookmarks(slots):
    """Load all bookmarks from config."""
    if not slots:
        slots = {}
    for slot_num in list(slots.keys()):
        if not osp.isfile(slots[slot_num][0]):
            slots.pop(slot_num)
    return slots


def load_bookmarks(filename, slots):
    """Load all bookmarks for a specific file from config."""
    bookmarks = _load_all_bookmarks(slots)
    return {k: v for k, v in bookmarks.items() if v[0] == filename}


def load_bookmarks_without_file(filename, slots):
    """Load all bookmarks but those from a specific file."""
    bookmarks = _load_all_bookmarks(slots)
    return {k: v for k, v in bookmarks.items() if v[0] != filename}


def update_bookmarks(filename, bookmarks, old_slots):
    """
    Compute an updated version of all the bookmarks from a specific file.

    Parameters
    ----------
    filename : str
        File path that the bookmarks are related too.
    bookmarks : dict
        New or changed bookmarks for the file.
    old_slots : dict
        Base general bookmarks entries available before any changes where done.

    Returns
    -------
    updated_slots : dict
        Updated general bookmarks.

    """
    if not osp.isfile(filename):
        return
    updated_slots = load_bookmarks_without_file(filename, old_slots)
    for slot_num, content in bookmarks.items():
        updated_slots[slot_num] = [filename, content[0], content[1]]
    return updated_slots



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""EditorSplitter Widget"""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

# Standard library imports
import logging

# Third party imports
import qstylizer.style
from qtpy import PYQT5, PYQT6
from qtpy.QtCore import QByteArray, Qt, Slot
from qtpy.QtWidgets import QSplitter

# Local imports
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.config.base import running_under_pytest
from spyder.plugins.editor.widgets.editorstack.editorstack import EditorStack
from spyder.py3compat import qbytearray_to_str
from spyder.utils.palette import SpyderPalette


logger = logging.getLogger(__name__)


class EditorSplitter(QSplitter, SpyderWidgetMixin):
    """QSplitter for editor windows."""

    CONF_SECTION = "editor"

    def __init__(self, parent, main_widget, menu_actions, first=False,
                 register_editorstack_cb=None, unregister_editorstack_cb=None,
                 use_switcher=True):
        """Create a splitter for dividing an editor window into panels.

        Adds a new EditorStack instance to this splitter.  If it's not
        the first splitter, clones the current EditorStack from the
        EditorMainWidget.

        Args:
            parent: Parent widget.
            main_widget: PluginMainWidget this widget belongs to.
            menu_actions: QActions to include from the parent.
            first: Boolean if this is the first splitter in the editor.
            register_editorstack_cb: Callback to register the EditorStack.
                        Defaults to main_widget.register_editorstack() to
                        register the EditorStack with the EditorMainWidget.
            unregister_editorstack_cb: Callback to unregister the EditorStack.
                        Defaults to main_widget.unregister_editorstack() to
                        unregister the EditorStack with the EditorMainWidget.
        """
        if PYQT5 or PYQT6:
            super().__init__(parent, class_parent=main_widget)
        else:
            QSplitter.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=main_widget)

        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setChildrenCollapsible(False)

        self.toolbar_list = None
        self.menu_list = None

        self.main_widget = main_widget

        if register_editorstack_cb is None:
            register_editorstack_cb = self.main_widget.register_editorstack
        self.register_editorstack_cb = register_editorstack_cb
        if unregister_editorstack_cb is None:
            unregister_editorstack_cb = self.main_widget.unregister_editorstack
        self.unregister_editorstack_cb = unregister_editorstack_cb

        self.menu_actions = menu_actions
        self.editorstack = EditorStack(self, menu_actions, use_switcher)
        self.register_editorstack_cb(self.editorstack)
        if not first:
            self.main_widget.clone_editorstack(editorstack=self.editorstack)
        self.editorstack.destroyed.connect(self.editorstack_closed)
        self.editorstack.sig_split_vertically.connect(
            lambda: self.split(orientation=Qt.Vertical))
        self.editorstack.sig_split_horizontally.connect(
            lambda: self.split(orientation=Qt.Horizontal))
        self.addWidget(self.editorstack)

        if not running_under_pytest():
            self.editorstack.set_color_scheme(main_widget._get_color_scheme())

        self.setStyleSheet(self._stylesheet)

    def closeEvent(self, event):
        """Override QWidget closeEvent().

        This event handler is called with the given event when Qt
        receives a window close request from a top-level widget.
        """
        QSplitter.closeEvent(self, event)

    def __give_focus_to_remaining_editor(self):
        focus_widget = self.main_widget.get_focus_widget()
        if focus_widget is not None:
            focus_widget.setFocus()

    @Slot()
    def editorstack_closed(self):
        logger.debug("Closing EditorStack")

        try:
            self.unregister_editorstack_cb(self.editorstack)
            self.editorstack = None
            close_splitter = self.count() == 1
            if close_splitter:
                # editorstack just closed was the last widget in this QSplitter
                self.close()
                return
            self.__give_focus_to_remaining_editor()
        except (RuntimeError, AttributeError):
            # editorsplitter has been destroyed (happens when closing a
            # EditorMainWindow instance)
            return

    def editorsplitter_closed(self):
        logger.debug("Closing EditorSplitter")
        try:
            close_splitter = self.count() == 1 and self.editorstack is None
        except RuntimeError:
            # editorsplitter has been destroyed (happens when closing a
            # EditorMainWindow instance)
            return

        if close_splitter:
            # editorsplitter just closed was the last widget in this QSplitter
            self.close()
            return
        elif self.count() == 2 and self.editorstack:
            # back to the initial state: a single editorstack instance,
            # as a single widget in this QSplitter: orientation may be changed
            self.editorstack.reset_orientation()

        self.__give_focus_to_remaining_editor()

    def split(self, orientation=Qt.Vertical):
        """
        Create and attach a new EditorSplitter to the current EditorSplitter.

        The new EditorSplitter widget will contain an EditorStack that
        is a clone of the current EditorStack.

        A single EditorSplitter instance can be split multiple times, but the
        orientation will be the same for all the direct splits.  If one of
        the child splits is split, then that split can have a different
        orientation.
        """
        logger.debug("Create a new EditorSplitter")
        self.setOrientation(orientation)
        self.editorstack.set_orientation(orientation)
        editorsplitter = EditorSplitter(
            self.parent(),
            self.main_widget,
            self.menu_actions,
            register_editorstack_cb=self.register_editorstack_cb,
            unregister_editorstack_cb=self.unregister_editorstack_cb
        )
        self.addWidget(editorsplitter)
        editorsplitter.destroyed.connect(self.editorsplitter_closed)
        current_editor = editorsplitter.editorstack.get_current_editor()
        if current_editor is not None:
            current_editor.setFocus()

    def iter_editorstacks(self):
        """Return the editor stacks for this splitter and every first child.

        Note: If a splitter contains more than one splitter as a direct
              child, only the first child's editor stack is included.

        Returns:
            List of tuples containing (EditorStack instance, orientation).
        """
        editorstacks = [(self.widget(0), self.orientation())]
        if self.count() > 1:
            editorsplitter = self.widget(1)
            editorstacks += editorsplitter.iter_editorstacks()
        return editorstacks

    def get_layout_settings(self):
        """Return the layout state for this splitter and its children.

        Record the current state, including file names and current line
        numbers, of the splitter panels.

        Returns:
            A dictionary containing keys {hexstate, sizes, splitsettings}.
                hexstate: String of saveState() for self.
                sizes: List for size() for self.
                splitsettings: List of tuples of the form
                       (orientation, cfname, clines) for each EditorSplitter
                       and its EditorStack.
                           orientation: orientation() for the editor
                                 splitter (which may be a child of self).
                           cfname: EditorStack current file name.
                           clines: Current line number for each file in the
                               EditorStack.
        """
        splitsettings = []
        for editorstack, orientation in self.iter_editorstacks():
            clines = []
            cfname = ''
            # XXX - this overrides value from the loop to always be False?
            orientation = False
            if hasattr(editorstack, 'data'):
                clines = [finfo.editor.get_cursor_line_number()
                          for finfo in editorstack.data]
                cfname = editorstack.get_current_filename()
            splitsettings.append((orientation == Qt.Vertical, cfname, clines))
        return dict(hexstate=qbytearray_to_str(self.saveState()),
                    sizes=self.sizes(), splitsettings=splitsettings)

    def set_layout_settings(self, settings, dont_goto=None):
        """Restore layout state for the splitter panels.

        Apply the settings to restore a saved layout within the editor.  If
        the splitsettings key doesn't exist, then return without restoring
        any settings.

        The current EditorSplitter (self) calls split() for each element
        in split_settings, thus recreating the splitter panels from the saved
        state.  split() also clones the editorstack, which is then
        iterated over to restore the saved line numbers on each file.

        The size and positioning of each splitter panel is restored from
        hexstate.

        Args:
            settings: A dictionary with keys {hexstate, sizes, orientation}
                    that define the layout for the EditorSplitter panels.
            dont_goto: Defaults to None, which positions the cursor to the
                    end of the editor.  If there's a value, positions the
                    cursor on the saved line number for each editor.
        """
        splitsettings = settings.get('splitsettings')
        if splitsettings is None:
            return
        splitter = self
        editor = None
        for i, (is_vertical, cfname, clines) in enumerate(splitsettings):
            if i > 0:
                splitter.split(Qt.Vertical if is_vertical else Qt.Horizontal)
                splitter = splitter.widget(1)
            editorstack = splitter.widget(0)
            for j, finfo in enumerate(editorstack.data):
                editor = finfo.editor
                # TODO: go_to_line is not working properly (the line it jumps
                # to is not the corresponding to that file). This will be fixed
                # in a future PR (which will fix spyder-ide/spyder#3857).
                if dont_goto is not None:
                    # Skip go to line for first file because is already there.
                    pass
                else:
                    try:
                        editor.go_to_line(clines[j])
                    except IndexError:
                        pass
        hexstate = settings.get('hexstate')
        if hexstate is not None:
            self.restoreState(
                QByteArray().fromHex(str(hexstate).encode('utf-8'))
            )
        sizes = settings.get('sizes')
        if sizes is not None:
            self.setSizes(sizes)
        if editor is not None:
            editor.clearFocus()
            editor.setFocus()

    @property
    def _stylesheet(self):
        css = qstylizer.style.StyleSheet()
        css.QSplitter.setValues(
            background=SpyderPalette.COLOR_BACKGROUND_1
        )
        return css.toString()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for editor.py
"""

# Standard library imports
import os
import os.path as osp
import sys
from unittest.mock import Mock, MagicMock

# Third party imports
import pytest
from flaky import flaky
from qtpy.QtCore import Qt
from qtpy.QtGui import QTextCursor
from qtpy.QtWidgets import QVBoxLayout, QWidget

# Local imports
from spyder.config.base import get_conf_path, running_in_ci
from spyder.plugins.editor.widgets.editorstack import EditorStack
from spyder.utils.stylesheet import APP_STYLESHEET
from spyder.widgets.findreplace import FindReplace


HERE = osp.abspath(osp.dirname(__file__))


# =============================================================================
# ---- Qt Test Fixtures
# =============================================================================
@pytest.fixture
def base_editor_bot(qtbot):
    editor_stack = EditorStack(None, [], False)
    editor_stack.set_find_widget(Mock())
    editor_stack.set_io_actions(Mock(), Mock(), Mock(), Mock())
    return editor_stack


@pytest.fixture
def editor_bot(base_editor_bot, mocker, qtbot):
    """
    Set up EditorStack with CodeEditor containing some Python code.
    The cursor is at the empty line below the code.
    The file in the editor is `foo.py` and has not been changed.
    Returns tuple with EditorStack and CodeEditor.
    """
    editor_stack = base_editor_bot
    text = ('a = 1\n'
            'print(a)\n'
            '\n'
            'x = 2')  # a newline is added at end
    finfo = editor_stack.new('foo.py', 'utf-8', text)
    finfo.newly_created = False
    editor_stack.autosave.file_hashes = {'foo.py': hash(text + '\n')}
    qtbot.addWidget(editor_stack)
    return editor_stack, finfo.editor


@pytest.fixture
def visible_editor_bot(editor_bot, mocker):
    """
    Set up an EditorStack to test functionalities that require a
    visible editor.
    """
    editorstack, editor = editor_bot

    # We need to patch osp.isfile to avoid the 'this file does not exist'
    # message box.
    mocker.patch(
        'spyder.plugins.editor.widgets.editorstack.editorstack.osp.isfile',
        returned_value=True
    )

    editorstack.show()

    return editorstack, editor


@pytest.fixture
def editor_find_replace_bot(base_editor_bot, qtbot):
    # Widget to show together the editor stack and findreplace ones together
    widget = QWidget()
    qtbot.addWidget(widget)
    widget.setStyleSheet(str(APP_STYLESHEET))

    # Widget's layout
    layout = QVBoxLayout()
    widget.setLayout(layout)

    editor_stack = base_editor_bot
    layout.addWidget(editor_stack)
    widget.editor_stack = editor_stack

    text = ('spam bacon\n'
            'spam sausage\n'
            'spam egg')
    finfo = editor_stack.new('spam.py', 'utf-8', text)
    widget.editor = finfo.editor

    find_replace = FindReplace(editor_stack, enable_replace=True)
    editor_stack.set_find_widget(find_replace)
    find_replace.set_editor(finfo.editor)
    widget.find_replace = find_replace
    layout.addWidget(find_replace)

    # Resize widget and show
    widget.resize(900, 360)
    widget.show()

    return widget


@pytest.fixture
def editor_cells_bot(base_editor_bot, qtbot):
    editor_stack = base_editor_bot
    text = ('# %%\n'
            '# 1 cell\n'
            '# print(1)\n'
            '# %%\n'
            '# 2 cell\n'
            '# print(2)\n'
            '# %%\n'
            '# 3 cell\n'
            '# print(3)\n')
    finfo = editor_stack.new('cells.py', 'utf-8', text)
    qtbot.addWidget(editor_stack)
    return editor_stack, finfo.editor


# =============================================================================
# ---- Tests
# =============================================================================
def test_scroll_line_up_and_down(visible_editor_bot, mocker, qtbot):
    """
    Test that the scroll line up and down functionalities are working
    as expected.
    """
    editorstack, editor = visible_editor_bot

    # Define text in the editor long enough to require a vertical scrollbar.
    editor.setPlainText('new_line\n' * 1000)
    editorstack.go_to_line(1)
    vsb = editor.verticalScrollBar()

    # Assert initial state.
    assert vsb.value() == 0
    assert vsb.maximum() > 0

    # Scroll line down five times.
    expected_vsb_value = 0
    for _ in range(5):
        expected_vsb_value += vsb.singleStep()
        editor.scroll_line_down()
        assert vsb.value() == expected_vsb_value

    # Scroll line up three times.
    for _ in range(3):
        expected_vsb_value += -vsb.singleStep()
        editor.scroll_line_up()
        assert vsb.value() == expected_vsb_value


def test_find_number_matches(editor_find_replace_bot):
    """Test for number matches in find/replace."""
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    finder.case_button.setChecked(True)
    text = ' test \nTEST \nTest \ntesT '
    editor.set_text(text)

    finder.search_text.add_text('test')
    finder.find(changed=False, forward=True, rehighlight=False,
                multiline_replace_check=False)
    editor_text = finder.number_matches_text.text()
    assert editor_text == '1 of 1'

    finder.search_text.add_text('fail')
    finder.find(changed=False, forward=True, rehighlight=False,
                multiline_replace_check=False)
    assert not finder.number_matches_text.isVisible()


def test_move_current_line_up(editor_bot):
    editor_stack, editor = editor_bot

    # Move second line up when nothing is selected.
    editor.go_to_line(2)
    editor.move_line_up()
    expected_new_text = ('print(a)\n'
                         'a = 1\n'
                         '\n'
                         'x = 2\n')
    assert editor.toPlainText() == expected_new_text

    # Move line up when already at the top.
    editor.move_line_up()
    assert editor.toPlainText() == expected_new_text

    # Move fourth line up when part of the line is selected.
    editor.go_to_line(4)
    editor.moveCursor(QTextCursor.Right, QTextCursor.MoveAnchor)
    for i in range(2):
        editor.moveCursor(QTextCursor.Right, QTextCursor.KeepAnchor)
    editor.move_line_up()
    expected_new_text = ('print(a)\n'
                         'a = 1\n'
                         'x = 2\n'
                         '\n')
    assert editor.toPlainText()[:] == expected_new_text


def test_move_current_line_down(editor_bot):
    editor_stack, editor = editor_bot

    # Move fourth line down when nothing is selected.
    editor.go_to_line(4)
    editor.move_line_down()
    expected_new_text = ('a = 1\n'
                         'print(a)\n'
                         '\n'
                         '\n'
                         'x = 2')
    assert editor.toPlainText() == expected_new_text

    # Move line down when already at the bottom.
    editor.move_line_down()
    assert editor.toPlainText() == expected_new_text

    # Move first line down when part of the line is selected.
    editor.go_to_line(1)
    editor.moveCursor(QTextCursor.Right, QTextCursor.MoveAnchor)
    for i in range(2):
        editor.moveCursor(QTextCursor.Right, QTextCursor.KeepAnchor)
    editor.move_line_down()
    expected_new_text = ('print(a)\n'
                         'a = 1\n'
                         '\n'
                         '\n'
                         'x = 2')
    assert editor.toPlainText() == expected_new_text


def test_move_multiple_lines_up(editor_bot):
    editor_stack, editor = editor_bot

    # Move second and third lines up.
    editor.go_to_line(2)
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Down, QTextCursor.KeepAnchor)
    cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
    editor.setTextCursor(cursor)
    editor.move_line_up()

    expected_new_text = ('print(a)\n'
                         '\n'
                         'a = 1\n'
                         'x = 2\n')
    assert editor.toPlainText() == expected_new_text

    # Move first and second lines up (to test already at top condition).
    editor.move_line_up()
    assert editor.toPlainText() == expected_new_text


@pytest.mark.skipif(running_in_ci() and os.name == 'nt',
                    reason="It fails on Windows CI")
def test_copy_lines_down_up(visible_editor_bot, qtbot):
    """
    Test that copy lines down and copy lines up are working as expected.
    """
    # Note that we need a visible editor because the copy lines down and copy
    # lines up functionalities both rely on a paint event override to work
    # as expected.

    editorstack, editor = visible_editor_bot

    # Assert initial state.
    editorstack.go_to_line(1)
    assert editor.get_cursor_line_column() == (0, 0)

    # Select some text.
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Down, QTextCursor.KeepAnchor)
    cursor.movePosition(QTextCursor.Down, QTextCursor.KeepAnchor)
    editor.setTextCursor(cursor)
    assert editor.get_cursor_line_column() == (2, 0)
    assert editor.textCursor().selection().toPlainText() == 'a = 1\nprint(a)\n'

    # Copy lines down.
    editor.duplicate_line_down()
    qtbot.wait(100)
    assert editor.get_cursor_line_column() == (4, 0)
    assert editor.textCursor().selection().toPlainText() == 'a = 1\nprint(a)\n'
    assert editor.toPlainText() == 'a = 1\nprint(a)\n' * 2 + '\nx = 2\n'

    # Copy lines up.
    editor.duplicate_line_up()
    qtbot.wait(100)
    assert editor.get_cursor_line_column() == (4, 0)
    assert editor.textCursor().selection().toPlainText() == 'a = 1\nprint(a)\n'
    assert editor.toPlainText() == 'a = 1\nprint(a)\n' * 3 + '\nx = 2\n'


def test_move_multiple_lines_down(editor_bot):
    editor_stack, editor = editor_bot

    # Move third and fourth lines down.
    editor.go_to_line(3)
    cursor = editor.textCursor()
    cursor.movePosition(QTextCursor.Down, QTextCursor.KeepAnchor)
    cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
    editor.setTextCursor(cursor)
    editor.move_line_down()

    expected_new_text = ('a = 1\n'
                         'print(a)\n'
                         '\n'
                         '\n'
                         'x = 2')
    assert editor.toPlainText() == expected_new_text

    # Move fourht and fifth lines down (to test already at bottom condition).
    editor.move_line_down()
    assert editor.toPlainText() == expected_new_text


def test_run_top_line(editor_bot, qtbot):
    editor_stack, editor = editor_bot
    editor.go_to_line(1) # line number is one based
    editor.move_cursor(3)
    text, _, _, _ = editor_stack.get_selection()
    editor_stack.advance_line()
    assert text == 'a = 1'

    # check cursor moves to start of next line; note line number is zero based
    assert editor.get_cursor_line_column() == (1, 0)


def test_run_last_nonempty_line(editor_bot, qtbot):
    editor_stack, editor = editor_bot
    editor.go_to_line(4)
    text, _, _, _ = editor_stack.get_selection()
    editor_stack.advance_line()
    assert text == 'x = 2'
    assert editor.get_cursor_line_column() == (4, 0) # check cursor moves down


def test_run_empty_line_in_middle(editor_bot, qtbot):
    editor_stack, editor = editor_bot
    editor.go_to_line(3)
    _, _, _, _ = editor_stack.get_selection()
    editor_stack.advance_line()
    assert editor.get_cursor_line_column() == (3, 0) # check cursor moves down


def test_run_last_line_when_empty(editor_bot, qtbot):
    editor_stack, editor = editor_bot
    _, _, _, _ = editor_stack.get_selection()
    editor_stack.advance_line()
    # check cursor doesn't move
    assert editor.get_cursor_line_column() == (4, 0)


def test_run_last_line_when_nonempty(editor_bot, qtbot):
    editor_stack, editor = editor_bot
    editor.stdkey_backspace() # delete empty line at end
    old_text = editor.toPlainText()
    text, _, _, _ = editor_stack.get_selection()
    editor_stack.advance_line()
    assert text == 'x = 2'
    expected_new_text = old_text + editor.get_line_separator()
    # check blank line got added
    assert editor.toPlainText() == expected_new_text
    assert editor.get_cursor_line_column() == (4, 0) # check cursor moves down


def test_find_replace_case_sensitive(editor_find_replace_bot):
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    finder.show(hide_replace=False)
    finder.case_button.setChecked(True)
    text = ' test \nTEST \nTest \ntesT '
    editor.set_text(text)
    finder.search_text.add_text('test')
    finder.replace_text.add_text('pass')
    finder.replace_find()
    finder.replace_find()
    finder.replace_find()
    finder.replace_find()
    editor_text = editor.toPlainText()
    assert editor_text == ' pass \nTEST \nTest \ntesT '


def test_replace_current_selected_line(editor_find_replace_bot, qtbot):
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    expected_new_text = ('ham bacon\n'
                         'spam sausage\n'
                         'spam egg')

    finder.show(hide_replace=False)
    qtbot.keyClicks(finder.search_text, 'spam')
    qtbot.keyClicks(finder.replace_text, 'ham')
    qtbot.keyPress(finder.replace_text, Qt.Key_Return)
    assert editor.toPlainText()[0:-1] == expected_new_text


@pytest.mark.skipif(sys.platform.startswith('linux'), reason="Fails in Linux")
def test_replace_enter_press(editor_find_replace_bot, qtbot):
    """Test advance forward pressing Enter, and backwards with Shift+Enter."""
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    text = '  \nspam \nspam \nspam '
    editor.set_text(text)
    finder.search_text.add_text('spam')
    finder.search_text.lineEdit().setFocus()

    # search forward
    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return)
    assert editor.get_cursor_line_column() == (1,4)

    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return)
    assert editor.get_cursor_line_column() == (2,4)

    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return)
    assert editor.get_cursor_line_column() == (3,4)

    # search backwards
    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return,
                   modifier=Qt.ShiftModifier)
    assert editor.get_cursor_line_column() == (2,4)

    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return,
                   modifier=Qt.ShiftModifier)
    assert editor.get_cursor_line_column() == (1,4)

    qtbot.keyClick(finder.search_text.lineEdit(), Qt.Key_Return,
                   modifier=Qt.ShiftModifier)
    assert editor.get_cursor_line_column() == (3,4)


def test_replace_plain_regex(editor_find_replace_bot, qtbot):
    """Test that regex reserved characters are displayed as plain text."""
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    expected_new_text = ('.\\[()]*test bacon\n'
                         'spam sausage\n'
                         'spam egg')
    finder.show(hide_replace=False)

    qtbot.keyClicks(finder.search_text, 'spam')
    qtbot.keyClicks(finder.replace_text, r'.\[()]*test')
    qtbot.keyPress(finder.replace_text, Qt.Key_Return)
    assert editor.toPlainText()[0:-1] == expected_new_text


def test_replace_invalid_regex(editor_find_replace_bot, qtbot):
    """Assert that replacing an invalid regexp does nothing."""
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    old_text = editor.toPlainText()
    finder.show(hide_replace=False)

    # Test with invalid search_text and valid replace_text
    qtbot.keyClicks(finder.search_text, '\\')
    qtbot.keyClicks(finder.replace_text, 'anything')

    if not finder.re_button.isChecked():
        qtbot.mouseClick(finder.re_button, Qt.LeftButton)

    qtbot.mouseClick(finder.replace_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text
    qtbot.mouseClick(finder.replace_sel_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text
    qtbot.mouseClick(finder.replace_all_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text

    # Test with valid search_text and invalid replace_text
    qtbot.keyClicks(finder.search_text, 'anything')
    qtbot.keyClicks(finder.replace_text, '\\')

    qtbot.mouseClick(finder.replace_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text
    qtbot.mouseClick(finder.replace_sel_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text
    qtbot.mouseClick(finder.replace_all_button, Qt.LeftButton)
    assert editor.toPlainText() == old_text


def test_replace_honouring_case(editor_find_replace_bot, qtbot):
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    expected_new_text = ('Spam bacon\n'
                         'Spam sausage\n'
                         'Spam egg\n'
                         'Spam potatoes')
    qtbot.keyClicks(editor, 'SpaM potatoes')

    finder.show(hide_replace=False)
    qtbot.keyClicks(finder.search_text, 'Spa[a-z]')
    qtbot.keyClicks(finder.replace_text, 'Spam')

    # Make sure regex button is set
    if not finder.re_button.isChecked():
        qtbot.mouseClick(finder.re_button, Qt.LeftButton)

    # Make sure case button is not set
    if finder.case_button.isChecked():
        qtbot.mouseClick(finder.case_button, Qt.LeftButton)

    # Replace all
    qtbot.mouseClick(finder.replace_all_button, Qt.LeftButton)
    assert editor.toPlainText() == expected_new_text


def test_selection_escape_characters(editor_find_replace_bot, qtbot):
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    expected_new_text = ('spam bacon\n'
                         'spam sausage\n'
                         'spam egg\n'
                         '\\n \\t some escape characters')
    qtbot.keyClicks(editor, '\\n \\t escape characters')

    finder.show(hide_replace=False)
    qtbot.keyClicks(finder.search_text, 'escape')
    qtbot.keyClicks(finder.replace_text, 'some escape')

    # Select last line
    cursor = editor.textCursor()
    cursor.select(QTextCursor.LineUnderCursor)
    assert cursor.selection().toPlainText() == "\\n \\t escape characters"

    # Replace
    finder.replace_find_selection()

    # Test that selection is correct
    assert cursor.selection().toPlainText() == "\\n \\t some escape characters"
    assert editor.toPlainText() == expected_new_text


def test_selection_backslash(editor_find_replace_bot, qtbot):
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    expected_new_text = ('spam bacon\n'
                         'spam sausage\n'
                         'spam egg\n'
                         r'a = r"\left\{" + "\\}\\right\n"')
    text_to_add = 'a = r"\\leeft\\{" + "\\\\}\\\\right\\n"'
    qtbot.keyClicks(editor, text_to_add)

    finder.show(hide_replace=False)
    qtbot.keyClicks(finder.search_text, 'leeft')
    qtbot.keyClicks(finder.replace_text, 'left')

    # Select last line
    cursor = editor.textCursor()
    cursor.select(QTextCursor.LineUnderCursor)
    assert cursor.selection().toPlainText() == text_to_add

    # Replace
    finder.replace_find_selection()
    assert editor.toPlainText() == expected_new_text


def test_advance_cell(editor_cells_bot):
    editor_stack, editor = editor_cells_bot

    # cursor at the end of the file
    assert editor.get_cursor_line_column() == (10, 0)

    # advance backwards to the beginning of the 3rd cell
    editor_stack.advance_cell(reverse=True)
    assert editor.get_cursor_line_column() == (6, 0)

    # advance backwards to 2nd cell
    editor_stack.advance_cell(reverse=True)
    assert editor.get_cursor_line_column() == (3, 0)
    # advance backwards to 1st cell
    editor_stack.advance_cell(reverse=True)
    assert editor.get_cursor_line_column() == (0, 0)

    # advance to 2nd cell
    editor_stack.advance_cell()
    assert editor.get_cursor_line_column() == (3, 0)
    # advance to 3rd cell
    editor_stack.advance_cell()
    assert editor.get_cursor_line_column() == (6, 0)


def test_get_current_word(base_editor_bot, qtbot):
    """Test getting selected valid python word."""
    editor_stack = base_editor_bot
    text = ('some words with non-ascii  characters\n'
            'niño\n'
            'garçon\n'
            'α alpha greek\n'
            '123valid_python_word')
    finfo = editor_stack.new('foo.py', 'utf-8', text)
    qtbot.addWidget(editor_stack)
    editor = finfo.editor
    editor.go_to_line(1)

    # Select some
    editor.moveCursor(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
    assert 'some' == editor.textCursor().selectedText()
    assert editor.get_current_word() == 'some'

    # Select niño
    editor.go_to_line(2)
    editor.moveCursor(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
    assert 'niño' == editor.textCursor().selectedText()
    assert editor.get_current_word() == 'niño'

    # Select garçon
    editor.go_to_line(3)
    editor.moveCursor(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
    assert 'garçon' == editor.textCursor().selectedText()
    assert editor.get_current_word() == 'garçon'

    # Select α
    editor.go_to_line(4)
    editor.moveCursor(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
    assert 'α' == editor.textCursor().selectedText()
    assert editor.get_current_word() == 'α'

    # Select valid_python_word, should search first valid python word
    editor.go_to_line(5)
    editor.moveCursor(QTextCursor.EndOfWord, QTextCursor.KeepAnchor)
    assert '123valid_python_word' == editor.textCursor().selectedText()
    assert editor.get_current_word() == 'valid_python_word'


def test_tab_keypress_properly_caught_find_replace(editor_find_replace_bot,
                                                   qtbot):
    """
    Test that tab works in find/replace dialog.

    Regression test for spyder-ide/spyder#3674.
    Mock test—more isolated but less flimsy.
    """
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    text = '  \nspam \nspam \nspam '
    editor.set_text(text)

    finder.focusNextChild = MagicMock(name="focusNextChild")
    qtbot.keyPress(finder.search_text, Qt.Key_Tab)
    finder.focusNextChild.assert_called_once_with()


@flaky(max_runs=3)
@pytest.mark.skipif(sys.platform.startswith('linux'), reason="Fails on Linux")
def test_tab_moves_focus_from_search_to_replace(editor_find_replace_bot,
                                                qtbot):
    """
    Test that tab works in find/replace dialog.

    Regression test for spyder-ide/spyder#3674.
    "Real world" test—more comprehensive but potentially less robust.
    """
    editor = editor_find_replace_bot.editor
    finder = editor_find_replace_bot.find_replace

    text = '  \nspam \nspam \nspam '
    editor.set_text(text)
    finder.show(hide_replace=False)

    qtbot.wait(100)
    finder.search_text.setFocus()
    qtbot.wait(100)
    assert finder.search_text.hasFocus()
    assert not finder.replace_text.hasFocus()
    qtbot.keyPress(finder.search_text, Qt.Key_Tab)
    qtbot.wait(100)
    assert not finder.search_text.hasFocus()
    assert finder.replace_text.hasFocus()


@flaky(max_runs=3)
@pytest.mark.skipif(running_in_ci(), reason="Fails on CIs")
def test_tab_copies_find_to_replace(editor_find_replace_bot, qtbot):
    """Check that text in the find box is copied to the replace box on tab
    keypress. Regression test spyder-ide/spyder#4482."""
    finder = editor_find_replace_bot.find_replace

    finder.show(hide_replace=False)
    finder.search_text.setFocus()
    finder.search_text.set_current_text('This is some test text!')
    qtbot.wait(500)
    qtbot.keyClick(finder.search_text, Qt.Key_Tab)
    assert finder.replace_text.currentText() == 'This is some test text!'


def test_update_matches_in_find_replace(editor_find_replace_bot, qtbot):
    """
    Check that the total number of matches in the FindReplace widget is updated
    when switching files.
    """
    editor_stack = editor_find_replace_bot.editor_stack
    finder = editor_find_replace_bot.find_replace

    # Search for "spam" in current file
    finder.show(hide_replace=False)
    finder.search_text.setFocus()
    finder.search_text.set_current_text('spam')
    qtbot.wait(500)
    qtbot.keyClick(finder.search_text, Qt.Key_Return)

    # Open a new file and only write "spam" on it
    editor_stack.new('foo.py', 'utf-8', 'spam')

    # Focus new file and check the number of matches and highlights was updated
    editor_stack.set_stack_index(1)
    codeeditor = editor_stack.get_current_editor()
    assert finder.number_matches_text.text() == '1 matches'
    assert len(codeeditor.decorations._decorations['find']) == 1
    qtbot.wait(500)

    # Focus initial file and check the number of matches and highlights was
    # updated
    editor_stack.set_stack_index(0)
    codeeditor = editor_stack.get_current_editor()
    assert len(codeeditor.decorations._decorations['find']) == 3
    qtbot.wait(500)
    assert finder.number_matches_text.text() == '3 matches'


def test_autosave_all(editor_bot, mocker):
    """
    Test that `autosave_all()` calls maybe_autosave() on all open buffers.

    The `editor_bot` fixture is constructed with one open file and the test
    opens another one with `new()`, so maybe_autosave should be called twice.
    """
    editor_stack, editor = editor_bot
    editor_stack.new('ham.py', 'utf-8', '')
    mocker.patch.object(editor_stack.autosave, 'maybe_autosave')
    editor_stack.autosave.autosave_all()
    expected_calls = [mocker.call(0), mocker.call(1)]
    actual_calls = editor_stack.autosave.maybe_autosave.call_args_list
    assert actual_calls == expected_calls


def test_maybe_autosave(editor_bot):
    """
    Test that maybe_autosave() saves text to correct autosave file if contents
    are changed.
    """
    editor_stack, editor = editor_bot
    editor.set_text('spam\n')
    editor_stack.autosave.maybe_autosave(0)
    autosave_filename = os.path.join(get_conf_path('autosave'), 'foo.py')
    assert open(autosave_filename).read() == 'spam\n'
    os.remove(autosave_filename)


def test_maybe_autosave_saves_only_if_changed(editor_bot, mocker):
    """
    Test that maybe_autosave() only saves text if text has changed.

    The `editor_bot` fixture creates a clean editor, so the first call to
    `maybe_autosave()` should not autosave. After call #2 we change the text,
    so call #2 should autosave. The text is not changed after call #2, so
    call #3 should not autosave.
    """
    editor_stack, editor = editor_bot
    mocker.patch.object(editor_stack, '_write_to_file')
    editor_stack.autosave.maybe_autosave(0)  # call #1, should not write
    assert editor_stack._write_to_file.call_count == 0
    editor.set_text('ham\n')
    editor_stack.autosave.maybe_autosave(0)  # call #2, should write
    assert editor_stack._write_to_file.call_count == 1
    editor_stack.autosave.maybe_autosave(0)  # call #3, should not write
    assert editor_stack._write_to_file.call_count == 1


def test_maybe_autosave_does_not_save_new_files(editor_bot, mocker):
    """Test that maybe_autosave() does not save newly created files."""
    editor_stack, editor = editor_bot
    editor_stack.data[0].newly_created = True
    mocker.patch.object(editor_stack, '_write_to_file')
    editor_stack.autosave.maybe_autosave(0)
    editor_stack._write_to_file.assert_not_called()


def test_opening_sets_file_hash(base_editor_bot, mocker):
    """Test that opening a file sets the file hash."""
    editor_stack = base_editor_bot
    mocker.patch('spyder.plugins.editor.widgets.editorstack.editorstack.encoding.read',
                 return_value=('my text', 42))
    filename = osp.realpath('/mock-filename')
    editor_stack.load(filename)
    expected = {filename: hash('my text')}
    assert editor_stack.autosave.file_hashes == expected


def test_reloading_updates_file_hash(base_editor_bot, mocker):
    """Test that reloading a file updates the file hash."""
    editor_stack = base_editor_bot
    mocker.patch('spyder.plugins.editor.widgets.editorstack.editorstack.encoding.read',
                 side_effect=[('my text', 42), ('new text', 42)])
    filename = osp.realpath('/mock-filename')
    finfo = editor_stack.load(filename)
    index = editor_stack.data.index(finfo)
    editor_stack.reload(index)
    expected = {filename: hash('new text')}
    assert editor_stack.autosave.file_hashes == expected


def test_closing_removes_file_hash(base_editor_bot, mocker):
    """Test that closing a file removes the file hash."""
    editor_stack = base_editor_bot
    mocker.patch('spyder.plugins.editor.widgets.editorstack.editorstack.encoding.read',
                 return_value=('my text', 42))
    filename = osp.realpath('/mock-filename')
    finfo = editor_stack.load(filename)
    index = editor_stack.data.index(finfo)
    editor_stack.close_file(index)
    assert editor_stack.autosave.file_hashes == {}


@pytest.mark.parametrize('filename', ['ham.py', 'ham.txt'])
def test_maybe_autosave_does_not_save_after_open(base_editor_bot, mocker,
                                                 qtbot, filename):
    """
    Test that maybe_autosave() does not save files immediately after opening.

    Files should only be autosaved after the user made changes.
    Editors use different highlighters depending on the filename, so we test
    both Python and text files. The latter covers spyder-ide/spyder#8654.
    """
    editor_stack = base_editor_bot
    mocker.patch('spyder.plugins.editor.widgets.editorstack.editorstack.encoding.read',
                 return_value=('spam\n', 42))
    editor_stack.load(filename)
    mocker.patch.object(editor_stack, '_write_to_file')
    qtbot.wait(100)  # Wait for PygmentsSH.makeCharlist() if applicable
    editor_stack.autosave.maybe_autosave(0)
    editor_stack._write_to_file.assert_not_called()


def test_maybe_autosave_does_not_save_after_reload(base_editor_bot, mocker):
    """
    Test that maybe_autosave() does not save files immediately after reloading.

    Spyder reloads the file if it has changed on disk. In that case, there is
    no need to autosave because the contents in Spyder are identical to the
    contents on disk.
    """
    editor_stack = base_editor_bot
    txt = 'spam\n'
    editor_stack.create_new_editor('ham.py', 'ascii', txt, set_current=True)
    mocker.patch.object(editor_stack, '_write_to_file')
    mocker.patch('spyder.plugins.editor.widgets.editorstack.editorstack.encoding.read',
                 return_value=(txt, 'ascii'))
    editor_stack.reload(0)
    editor_stack.autosave.maybe_autosave(0)
    editor_stack._write_to_file.assert_not_called()

def test_autosave_updates_name_mapping(editor_bot, mocker, qtbot):
    """Test that maybe_autosave() updates name_mapping."""
    editor_stack, editor = editor_bot
    assert editor_stack.autosave.name_mapping == {}
    mocker.patch.object(editor_stack, '_write_to_file')
    editor.set_text('spam\n')
    editor_stack.autosave.maybe_autosave(0)
    expected = {'foo.py': os.path.join(get_conf_path('autosave'), 'foo.py')}
    assert editor_stack.autosave.name_mapping == expected


def test_maybe_autosave_handles_error(editor_bot, mocker):
    """Test that autosave() ignores errors when writing to file."""
    editor_stack, editor = editor_bot
    mock_write = mocker.patch.object(editor_stack, '_write_to_file')
    mock_dialog = mocker.patch(
        'spyder.plugins.editor.utils.autosave.AutosaveErrorDialog')
    try:
        mock_write.side_effect = PermissionError
    except NameError:  # Python 2
        mock_write.side_effect = IOError
    editor.set_text('spam\n')
    editor_stack.autosave.maybe_autosave(0)
    assert mock_dialog.called


def test_remove_autosave_file(editor_bot, mocker, qtbot):
    """
    Test that remove_autosave_file() removes the autosave file.

    Also, test that it updates `name_mapping`.
    """
    editor_stack, editor = editor_bot
    autosave = editor_stack.autosave
    editor.set_text('spam\n')

    autosave.maybe_autosave(0)

    autosave_filename = os.path.join(get_conf_path('autosave'), 'foo.py')
    assert os.access(autosave_filename, os.R_OK)
    expected = {'foo.py': autosave_filename}
    assert autosave.name_mapping == expected

    autosave.remove_autosave_file(editor_stack.data[0].filename)

    assert not os.access(autosave_filename, os.R_OK)
    assert autosave.name_mapping == {}


def test_ipython_files(base_editor_bot, qtbot):
    """Test support for IPython files in the editor."""
    # Load IPython file
    editor_stack = base_editor_bot
    editor_stack.load(osp.join(HERE, 'assets', 'ipython_file.ipy'))

    editor = editor_stack.get_current_editor()
    editor.completions_available = True

    # Assert file is recognized as IPython
    assert editor.is_ipython()

    # Assert we transform IPython cells to valid Python code when opening
    # the file
    with qtbot.waitSignal(editor.sig_perform_completion_request) as blocker:
        editor.document_did_open()

    request_params = blocker.args[2]
    assert 'get_ipython' in request_params['text']

    # Assert we transform IPython cells to valid Python code when modifying
    # the file
    with qtbot.waitSignal(editor.sig_perform_completion_request) as blocker:
        editor.document_did_change()

    params = blocker.args[2]
    assert 'get_ipython' in params['text']

    # Mock linting results for this file. This is actually what's returned by
    # Pyflakes.
    editor._diagnostics = [
        {
            'source': 'pyflakes',
            'range': {
                'start': {'line': 8, 'character': 0},
                'end': {'line': 8, 'character': 19}
            },
            'message': "'numpy as np' imported but unused", 'severity': 2
        },
        {
            'source': 'pyflakes',
            'range': {
                'start': {'line': 11, 'character': 0},
                'end': {'line': 11, 'character': 47}
            },
            'message': "undefined name 'get_ipython'", 'severity': 1
        }
    ]

    # Setting them in the editor
    editor.set_errors()

    # Measure how many blocks have attached linting data to them
    blocks_with_data = 0
    block = editor.document().firstBlock()
    while block.isValid():
        data = block.userData()
        if data and data.code_analysis:
            blocks_with_data += 1
        block = block.next()

    # Assert we omitted the get_ipython message from linting
    assert blocks_with_data == 1


if __name__ == "__main__":
    pytest.main(['test_editor.py', '-vv', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for pyenv.py"""

import sys
import time

import pytest

from spyder.config.base import running_in_ci
from spyder.utils.programs import find_program
from spyder.utils.pyenv import get_list_pyenv_envs, get_list_pyenv_envs_cache


if not find_program('pyenv'):
    pytest.skip("Requires pyenv to be installed", allow_module_level=True)


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
@pytest.mark.skipif(not sys.platform.startswith('linux'),
                    reason="Only runs on Linux")
def test_get_list_pyenv_envs():
    output = get_list_pyenv_envs()
    expected_envs = ['Pyenv: 3.8.1']
    assert set(expected_envs) == set(output.keys())


@pytest.mark.skipif(not running_in_ci(), reason="Only meant for CIs")
@pytest.mark.skipif(not sys.platform.startswith('linux'),
                    reason="Only runs on Linux")
def test_get_list_pyenv_envs_cache():
    time0 = time.time()
    output = get_list_pyenv_envs_cache()
    time1 = time.time()

    assert output != {}
    assert (time1 - time0) < 0.01



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder run models."""

# Standard library imports
from collections import OrderedDict
from typing import Dict, Tuple, Optional, Set, Union, List

# Third-party imports
from qtpy.QtCore import (Qt, Signal, QAbstractListModel, QModelIndex,
                         QAbstractTableModel)

# Local imports
from spyder.api.translations import _
from spyder.plugins.run.api import (
    ExtendedRunExecutionParameters,
    RunConfigurationMetadata,
    RunContext,
    StoredRunConfigurationExecutor,
    StoredRunExecutorParameters,
    SupportedExecutionRunConfiguration,
)


class RunExecutorListModel(QAbstractListModel):
    sig_executor_widget_changed = Signal(object)

    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.uuid: Optional[str] = None
        self.current_input: Optional[Tuple[str, str]] = None
        self.executor_names: Dict[str, str] = {}
        self.executor_configurations: Dict[Tuple[str, str], Dict[
            str, SupportedExecutionRunConfiguration]] = {}
        self.executors_per_input: Dict[Tuple[str, str], Dict[str, int]] = {}
        self.inverted_pos: Dict[Tuple[str, str], Dict[int, str]] = {}
        self.executor_priority: Dict[Tuple[str, str], Dict[str, int]] = {}

    def set_executor_name(self, executor_id: str, executor_name: str):
        self.executor_names[executor_id] = executor_name

    def add_input_executor_configuration(
        self,
        ext: str,
        context_id: str,
        executor_id: str,
        config: SupportedExecutionRunConfiguration
    ):
        executor_conf = self.executor_configurations.get((ext, context_id), {})
        executor_conf[executor_id] = config
        self.executor_configurations[(ext, context_id)] = executor_conf

        input_executors = self.executors_per_input.get((ext, context_id), {})
        all_exec_prio = self.executor_priority.get((ext, context_id), {})
        priority = config['priority']

        # Remove if existing
        input_executors.pop(executor_id, None)
        all_exec_prio.pop(executor_id, None)
        all_exec_prio[executor_id] = priority

        input_values = list(all_exec_prio.items())
        input_values = sorted(input_values, key=lambda k: k[1])

        input_values = [(x, i) for i, (x, _) in enumerate(input_values)]
        self.executors_per_input[(ext, context_id)] = dict(input_values)

        self.inverted_pos[(ext, context_id)] = {
            v: k for (k, v) in input_values}
        self.executor_priority[(ext, context_id)] = all_exec_prio

    def remove_input_executor_configuration(
        self,
        ext: str,
        context_id: str,
        executor_id: str
    ):
        input_executors = self.executors_per_input[(ext, context_id)]
        pos = input_executors.pop(executor_id)

        inverted_executors = self.inverted_pos[(ext, context_id)]
        inverted_executors.pop(pos)

    def switch_input(self, uuid: str, run_input: Tuple[str, str]):
        if run_input in self.executors_per_input:
            self.beginResetModel()
            self.current_input = run_input
            self.uuid = uuid
            self.endResetModel()
        else:
            raise ValueError(
                f'The requested run input combination {run_input} is not '
                f'registered'
            )

    def get_selected_run_executor(
        self,
        index: int
    ) -> Tuple[str, SupportedExecutionRunConfiguration]:

        executor_indices = self.inverted_pos[self.current_input]
        executor_name = executor_indices[index]
        executors = self.executor_configurations[self.current_input]
        executor = executors[executor_name]

        return executor_name, executor

    def get_run_executor_index(self, executor_name: str) -> int:
        ordered_executors = self.executors_per_input[self.current_input]
        executor_index = ordered_executors[executor_name]
        return executor_index

    def get_initial_index(self) -> int:
        last_executor = self.parent.get_last_used_executor_parameters(
            self.uuid)
        executor_id = last_executor['executor']
        pos = 0

        if executor_id is not None:
            executors = self.executors_per_input[self.current_input]
            pos = executors[executor_id]

        return pos

    def get_default_executor(self, input: Tuple[str, str]) -> str:
        executors = self.inverted_pos[input]
        return executors[0]

    def executor_supports_configuration(
        self,
        executor: str,
        exec_input: Tuple[str, str]
    ) -> bool:

        input_executors = self.executor_configurations.get(exec_input, {})
        return executor in input_executors

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            executor_indices = self.inverted_pos[self.current_input]
            executor_id = executor_indices[index.row()]
            return self.executor_names[executor_id]

    def rowCount(self, parent: QModelIndex = None) -> int:
        executors = self.executors_per_input.get(self.current_input, {})
        return len(executors)

    def __contains__(self, exec_input: Tuple[str, str]) -> bool:
        return exec_input in self.executor_configurations

    def __len__(self):
        return len(self.executor_configurations)

    def __iter__(self):
        return iter(self.executor_configurations)

    def __getitem__(
            self, input_executor: tuple) -> SupportedExecutionRunConfiguration:
        (input, executor) = input_executor
        input_executors = self.executor_configurations[input]
        return input_executors[executor]


class RunConfigurationListModel(QAbstractListModel):
    def __init__(self, parent, executor_model):
        super().__init__(parent)
        self.parent = parent
        self.selected_metadata: Optional[RunConfigurationMetadata] = None
        self.current_configuration: Optional[str] = None
        self.metadata_index: Dict[int, str] = {}
        self.inverted_index: Dict[str, int] = {}
        self.run_configurations: OrderedDict[
            str, RunConfigurationMetadata] = OrderedDict()
        self.executor_model: RunExecutorListModel = executor_model

    def get_metadata_context_extension(self, uuid: str):
        run_conf = self.run_configurations[uuid]
        return run_conf['context'], run_conf['input_extension']

    def set_current_run_configuration(self, uuid: str):
        self.current_configuration = uuid
    
    def get_current_run_configuration(self):
        return self.current_configuration

    def get_initial_index(self) -> int:
        return self.inverted_index[self.current_configuration]

    def get_selected_metadata(self) -> Optional[RunConfigurationMetadata]:
        return self.selected_metadata

    def get_metadata(self, index: int) -> RunConfigurationMetadata:
        uuid = self.metadata_index[index]
        metadata = self.run_configurations[uuid]
        return metadata

    def update_index(self, index: int):
        uuid = self.metadata_index[index]
        metadata = self.run_configurations[uuid]
        context_name = metadata['context']['name']
        context_id = getattr(RunContext, context_name)
        ext = metadata['input_extension']
        self.selected_metadata = metadata
        self.executor_model.switch_input(uuid, (ext, context_id))

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        if role == Qt.DisplayRole or role == Qt.EditRole:
            uuid = self.metadata_index[index.row()]
            metadata = self.run_configurations[uuid]
            return metadata['name']

    def rowCount(self, parent: QModelIndex = None) -> int:
        return len(self.run_configurations)

    def get_run_configuration_parameters(
        self,
        uuid: str,
        executor: str
    ) -> Optional[StoredRunExecutorParameters]:

        context, ext = self.get_metadata_context_extension(uuid)
        context_name = context['name']
        context_id = getattr(RunContext, context_name)

        return self.parent.get_executor_configuration_parameters(
            executor, ext, context_id)

    def get_last_used_execution_params(
        self,
        uuid: str,
        executor: str
    ) -> Optional[str]:

        return self.parent.get_last_used_execution_params(uuid, executor)

    def get_last_used_executor_parameters(
        self,
        uuid: str
    ) -> StoredRunConfigurationExecutor:

        return self.parent.get_last_used_executor_parameters(uuid)

    def pop(self, uuid: str) -> RunConfigurationMetadata:
        item = self.run_configurations.pop(uuid)
        self.metadata_index = dict(enumerate(self.run_configurations))
        self.inverted_index = {v: k for k, v in self.metadata_index.items()}
        if self.current_configuration not in self.inverted_index:
            self.current_configuration = None
        self.dataChanged.emit(self.createIndex(0, 0),
                              self.createIndex(len(self.metadata_index), 0))
        return item

    def __iter__(self):
        return iter(self.run_configurations)

    def __len__(self):
        return len(self.run_configurations)

    def __getitem__(self, uuid: str) -> RunConfigurationMetadata:
        return self.run_configurations[uuid]

    def __setitem__(self, uuid: str, metadata: RunConfigurationMetadata):
        self.run_configurations[uuid] = metadata
        self.metadata_index[len(self.metadata_index)] = uuid
        self.inverted_index[uuid] = len(self.inverted_index)
        self.dataChanged.emit(self.createIndex(0, 0),
                              self.createIndex(len(self.metadata_index), 0))

    def __contains__(self, uuid: str):
        return uuid in self.run_configurations


class RunExecutorParameters(QAbstractListModel):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.executor_conf_params: Dict[
            str, ExtendedRunExecutionParameters] = {}
        self.params_index: Dict[int, str] = {}
        self.inverse_index: Dict[str, int] = {}

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        pos = index.row()
        pos = 0 if pos == -1 else pos
        params_id = self.params_index[pos]
        params = self.executor_conf_params[params_id]
        params_name = params['name']
        if role == Qt.DisplayRole or role == Qt.EditRole:
            return params_name

    def rowCount(self, parent: QModelIndex = ...) -> int:
        return len(self.executor_conf_params)

    def get_parameters(self, index: int) -> ExtendedRunExecutionParameters:
        params_id = self.params_index[index]
        return self.executor_conf_params[params_id]

    def get_parameters_uuid_name(
        self, index: int
    ) -> Tuple[Optional[str], Optional[str]]:

        params_id = self.params_index[index]
        params = self.executor_conf_params[params_id]
        return params['uuid'], params['name']

    def set_parameters(
        self,
        parameters: Dict[str, ExtendedRunExecutionParameters]
    ):
        self.beginResetModel()
        self.executor_conf_params = parameters
        self.params_index = dict(enumerate(self.executor_conf_params))
        self.inverse_index = {
            self.params_index[k]: k for k in self.params_index
        }
        self.endResetModel()

    def get_parameters_index_by_uuid(
        self, parameters_uuid: Optional[str]
    ) -> int:
        return self.inverse_index.get(parameters_uuid, 0)

    def get_parameters_index_by_name(self, parameters_name: str) -> int:
        index = -1
        for id_, idx in self.inverse_index.items():
            if self.executor_conf_params[id_]['name'] == parameters_name:
                index = idx
                break

        return index

    def get_parameter_names(self, filter_global: bool = False) -> List[str]:
        """
        Get all parameter names for this executor.

        Parameters
        ----------
        filter_global: bool, optional
            Whether to filter global parameters from the results.
        """
        names = []
        for params in self.executor_conf_params.values():
            if filter_global:
                if params["file_uuid"] is not None:
                    names.append(params["name"])
            else:
                names.append(params["name"])

        return names

    def get_number_of_custom_params(self, global_params_name: str) -> int:
        """
        Get the number of custom parameters derived from a set of global ones.

        Parameters
        ----------
        global_params_name: str
            Name of the global parameters.
        """
        names = self.get_parameter_names(filter_global=True)
        return len(
            [name for name in names if name.startswith(global_params_name)]
        )

    def __len__(self) -> int:
        return len(self.executor_conf_params)


class RunExecutorNamesListModel(QAbstractListModel):
    def __init__(self, parent, executor_model: RunExecutorListModel):
        super().__init__(parent)
        self.executor_model = executor_model
        self.executor_configurations: OrderedDict[
            str, Set[Tuple[str, str]]] = OrderedDict({})

        for input_conf in self.executor_model.executors_per_input:
            executors_available = self.executor_model.executors_per_input[
                input_conf]
            for executor in executors_available:
                exec_input_conf = self.executor_model[(input_conf, executor)]
                ConfigWidget = exec_input_conf.get(
                    'configuration_widget', None)
                if ConfigWidget is not None:
                    input_set = self.executor_configurations.get(
                        executor,set({}))
                    input_set |= {input_conf}
                    self.executor_configurations[executor] = input_set

        self.executor_indexed_list: Dict[int, str] = dict(
            enumerate(self.executor_configurations))

        self.executor_list: Dict[str, int] = {
            v: k for k, v in self.executor_indexed_list.items()}

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> str:
        row = index.row()
        if role == Qt.DisplayRole or role == Qt.EditRole:
            executor_id = self.executor_indexed_list[row]
            return self.executor_model.executor_names[executor_id]

    def rowCount(self, parent: QModelIndex = None) -> int:
        return len(self.executor_list)

    def selected_executor(
            self, index: int) -> Tuple[str, Set[Tuple[str, str]]]:
        executor_name = self.executor_indexed_list[index]
        return executor_name, self.executor_configurations[executor_name]


class ExecutorRunParametersTableModel(QAbstractTableModel):
    NAME = 0
    EXTENSION = 1
    CONTEXT = 2

    sig_data_changed = Signal()

    def __init__(self, parent):
        super().__init__(parent)
        self.executor_conf_params: Dict[
            Tuple[str, str, str], ExtendedRunExecutionParameters
        ] = {}

        self.params_index: Dict[int, Tuple[str, str, str]] = {}
        self.inverse_index: Dict[Tuple[str, str, str], int] = {}

    def data(self, index: QModelIndex,
             role: int = Qt.DisplayRole) -> Union[str, int]:
        row = index.row()
        column = index.column()

        params_idx = self.params_index[row]
        (extension, context, __) = params_idx
        params = self.executor_conf_params[params_idx]

        if role == Qt.DisplayRole or role == Qt.EditRole:
            if column == self.EXTENSION:
                return extension
            elif column == self.CONTEXT:
                return context
            elif column == self.NAME:
                return _("Default") if params['default'] else params['name']
        elif role == Qt.TextAlignmentRole:
            return int(Qt.AlignHCenter | Qt.AlignVCenter)
        elif role == Qt.ToolTipRole:
            return _("Double-click to view or edit")

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.DisplayRole
    ) -> Union[str, int]:

        if role == Qt.TextAlignmentRole:
            if orientation == Qt.Horizontal:
                return int(Qt.AlignHCenter | Qt.AlignVCenter)
            return int(Qt.AlignRight | Qt.AlignVCenter)

        if role == Qt.DisplayRole or role == Qt.EditRole:
            if orientation == Qt.Horizontal:
                if section == self.EXTENSION:
                    return _('File extension')
                elif section == self.CONTEXT:
                    return _('Context')
                elif section == self.NAME:
                    return _('Name')

    def rowCount(self, parent: QModelIndex = None) -> int:
        return len(self.params_index)

    def columnCount(self, parent: QModelIndex = None) -> int:
        return 3

    def set_parameters(
        self,
        params: Dict[Tuple[str, str, str], ExtendedRunExecutionParameters]
    ):
        self.beginResetModel()

        # Reorder params so that Python and IPython extensions are shown first
        # and second by default, respectively.
        ordered_params = []
        for k, v in params.items():
            if k[0] == "py":
                ordered_params.insert(0, (k, v))
            elif k[0] == "ipy":
                ordered_params.insert(1, (k, v))
            else:
                ordered_params.append((k, v))
        params = dict(ordered_params)

        # Update attributes
        self.executor_conf_params = params
        self.params_index = dict(enumerate(params))
        self.inverse_index = {v: k for k, v in self.params_index.items()}

        self.endResetModel()

    def get_current_view(
        self
    ) -> Dict[Tuple[str, str, str], ExtendedRunExecutionParameters]:

        return self.executor_conf_params

    def get_tuple_index(self, index: int) -> Tuple[str, str, str]:
        return self.params_index[index]

    def reset_model(self):
        self.beginResetModel()
        self.endResetModel()

    def get_parameters_name(self, idx: Tuple[str, str, str]) -> str:
        params = self.executor_conf_params[idx]
        return params['name']

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        self.beginResetModel()
        reverse = order == Qt.DescendingOrder

        if column == self.EXTENSION:
            sorted_keys = sorted(self.executor_conf_params,
                                 key=lambda x: x[0],
                                 reverse=reverse)
        elif column == self.CONTEXT:
            sorted_keys = sorted(self.executor_conf_params,
                                 key=lambda x: x[1],
                                 reverse=reverse)
        elif column == self.NAME:
            sorted_keys = sorted(
                self.executor_conf_params,
                key=lambda x: self.executor_conf_params[x]['name'],
                reverse=reverse)

        self.params_index = dict(enumerate(sorted_keys))
        self.inverse_index = {v: k for k, v in self.params_index.items()}
        self.endResetModel()

    def apply_changes(
        self,
        changes: List[Tuple[str, ExtendedRunExecutionParameters]],
        extension: str,
        context: str
    ):
        self.beginResetModel()

        for (operation, params) in changes:
            key = (extension, context, params['uuid'])
            if operation == 'deleted':
                self.executor_conf_params.pop(key)
            else:
                self.executor_conf_params[key] = params

        self.params_index = dict(enumerate(self.executor_conf_params))
        self.inverse_index = {v: k for k, v in self.params_index.items()}

        self.endResetModel()

        self.sig_data_changed.emit()

    def get_parameter_names(self) -> Dict[Tuple[str, str], List[str]]:
        """Get all parameter names per extension and context."""
        names = {}
        for k, v in self.executor_conf_params.items():
            extension_context = (k[0], k[1])
            current_names = names.get(extension_context, [])
            current_names.append(_("Default") if v["default"] else v["name"])
            names[extension_context] = current_names

        return names

    def __len__(self):
        return len(self.inverse_index)

    def __iter__(self):
        return iter(self.executor_conf_params)

    def __getitem__(
        self,
        index: int
    ) -> Tuple[str, str, ExtendedRunExecutionParameters]:

        tuple_index = self.params_index[index]
        (ext, context, __) = tuple_index
        return (ext, context, self.executor_conf_params[tuple_index])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Working Directory Plugin"""

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

# Third party imports
from qtpy.QtWidgets import (QButtonGroup, QGroupBox, QHBoxLayout, QLabel,
                            QVBoxLayout)

# Local imports
from spyder.config.base import _
from spyder.api.preferences import PluginConfigPage
from spyder.utils.misc import getcwd_or_home


class WorkingDirectoryConfigPage(PluginConfigPage):

    def setup_page(self):
        about_label = QLabel(
            _("This is the directory that will be set as the default for "
              "the IPython console and Files panes.")
        )
        about_label.setWordWrap(True)

        # Startup directory
        startup_group = QGroupBox(_("Startup"))
        startup_bg = QButtonGroup(startup_group)
        startup_label = QLabel(
            _("At startup, the working directory is:")
        )
        startup_label.setWordWrap(True)
        lastdir_radio = self.create_radiobutton(
            _("The project (if open) or user home directory"),
            'startup/use_project_or_home_directory',
            tip=_("The startup working dir will be root of the "
                  "current project if one is open, otherwise the "
                  "user home directory"),
            button_group=startup_bg
        )
        thisdir_radio = self.create_radiobutton(
            _("The following directory:"),
            'startup/use_fixed_directory',
            _("At startup, the current working directory will be the "
              "specified path"),
            button_group=startup_bg
        )
        thisdir_bd = self.create_browsedir(
            "",
            'startup/fixed_directory',
            getcwd_or_home()
        )
        thisdir_radio.radiobutton.toggled.connect(thisdir_bd.setEnabled)
        lastdir_radio.radiobutton.toggled.connect(thisdir_bd.setDisabled)
        thisdir_layout = QHBoxLayout()
        thisdir_layout.addWidget(thisdir_radio)
        thisdir_layout.addWidget(thisdir_bd)

        startup_layout = QVBoxLayout()
        startup_layout.addWidget(startup_label)
        startup_layout.addWidget(lastdir_radio)
        startup_layout.addLayout(thisdir_layout)
        startup_group.setLayout(startup_layout)

        # Console Directory
        console_group = QGroupBox(_("New consoles"))
        console_label = QLabel(
            _("The working directory for new IPython consoles is:")
        )
        console_label.setWordWrap(True)
        console_bg = QButtonGroup(console_group)
        console_project_radio = self.create_radiobutton(
            _("The project (if open) or user home directory"),
            'console/use_project_or_home_directory',
            tip=_("The working dir for new consoles will be root of the "
                  "project if one is open, otherwise the user home directory"),
            button_group=console_bg
        )
        console_cwd_radio = self.create_radiobutton(
            _("The working directory of the current console"),
            'console/use_cwd',
            button_group=console_bg
        )
        console_dir_radio = self.create_radiobutton(
            _("The following directory:"),
            'console/use_fixed_directory',
            _("The directory when a new console is open will be the "
              "specified path"),
            button_group=console_bg
        )
        console_dir_bd = self.create_browsedir(
            "",
            'console/fixed_directory',
            getcwd_or_home()
        )
        console_dir_radio.radiobutton.toggled.connect(console_dir_bd.setEnabled)
        console_project_radio.radiobutton.toggled.connect(console_dir_bd.setDisabled)
        console_cwd_radio.radiobutton.toggled.connect(console_dir_bd.setDisabled)
        console_dir_layout = QHBoxLayout()
        console_dir_layout.addWidget(console_dir_radio)
        console_dir_layout.addWidget(console_dir_bd)

        console_layout = QVBoxLayout()
        console_layout.addWidget(console_label)
        console_layout.addWidget(console_project_radio)
        console_layout.addWidget(console_cwd_radio)
        console_layout.addLayout(console_dir_layout)
        console_group.setLayout(console_layout)

        vlayout = QVBoxLayout()
        vlayout.addWidget(about_label)
        vlayout.addSpacing(10)
        vlayout.addWidget(startup_group)
        vlayout.addWidget(console_group)
        vlayout.addStretch(1)
        self.setLayout(vlayout)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""
Tests for namespacebrowser.py
"""

# Standard library imports
import string
from unittest.mock import Mock, patch

# Third party imports
from flaky import flaky
import pytest
from qtpy.QtCore import Qt, QPoint, QModelIndex

# Local imports
from spyder.plugins.variableexplorer.widgets.namespacebrowser import (
    NamespaceBrowser)
from spyder.widgets.collectionseditor import ROWS_TO_LOAD
from spyder.widgets.tests.test_collectioneditor import data, data_table


# =============================================================================
# ---- Fixtures
# =============================================================================
@pytest.fixture
def namespacebrowser(qtbot):
    browser = NamespaceBrowser(None)
    browser.set_shellwidget(Mock())
    browser.setup()
    browser.resize(640, 480)
    browser.show()
    qtbot.addWidget(browser)
    return browser


# =============================================================================
# ---- Tests
# =============================================================================
@flaky(max_runs=5)
def test_automatic_column_width(namespacebrowser):
    browser = namespacebrowser

    col_width = [browser.editor.columnWidth(i) for i in range(4)]
    browser.set_data({'a_variable':
        {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
         'numpy_type': 'Unknown'}})
    new_col_width = [browser.editor.columnWidth(i) for i in range(4)]
    assert browser.editor.automatic_column_width
    assert col_width != new_col_width  # Automatic col width is on
    browser.editor.horizontalHeader()._handle_section_is_pressed = True
    browser.editor.setColumnWidth(0, 100)  # Simulate user changing col width
    assert browser.editor.automatic_column_width == False
    browser.set_data({'a_lengthy_variable_name_which_should_change_width':
        {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
         'numpy_type': 'Unknown'}})
    assert browser.editor.columnWidth(0) == 100  # Automatic col width is off


def test_sort_by_column(namespacebrowser, qtbot):
    """
    Test that clicking the header view the namespacebrowser is sorted.
    Regression test for spyder-ide/spyder#9835 .
    """
    browser = namespacebrowser

    browser.set_data(
        {'a_variable':
            {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
             'numpy_type': 'Unknown'},
         'b_variable':
            {'type': 'int', 'size': 1, 'view': '2', 'python_type': 'int',
             'numpy_type': 'Unknown'}}
    )

    header = browser.editor.horizontalHeader()

    # Check header is clickable
    assert header.sectionsClickable()

    model = browser.editor.model

    # Base check of the model
    assert model.rowCount() == 2
    assert model.columnCount() == 5
    assert data_table(model, 2, 4) == [['a_variable', 'b_variable'],
                                       ['int', 'int'],
                                       [1, 1],
                                       ['1', '2']]

    with qtbot.waitSignal(header.sectionClicked):
        browser.show()
        qtbot.mouseClick(header.viewport(), Qt.LeftButton, pos=QPoint(1, 1))

    # Check sort effect
    assert data_table(model, 2, 4) == [['b_variable', 'a_variable'],
                                       ['int', 'int'],
                                       [1, 1],
                                       ['2', '1']]


def test_keys_sorted_and_sort_with_large_rows(namespacebrowser, qtbot):
    """
    Test that keys are sorted and sorting works as expected when
    there's a large number of rows.

    This is a regression test for issue spyder-ide/spyder#10702
    """
    browser = namespacebrowser

    # Create variables.
    variables = {}
    variables['i'] = (
        {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
         'numpy_type': 'Unknown'}
    )

    for i in range(100):
        if i < 10:
            var = 'd_0' + str(i)
        else:
            var = 'd_' + str(i)
        variables[var] = (
            {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
             'numpy_type': 'Unknown'}
        )

    # Set data
    browser.set_data(variables)

    # Assert we loaded the expected amount of data and that we can fetch
    # more.
    model = browser.editor.model
    assert model.rowCount() == ROWS_TO_LOAD
    assert model.canFetchMore(QModelIndex())

    # Assert keys are sorted
    assert data(model, 49, 0) == 'd_49'

    # Sort
    header = browser.editor.horizontalHeader()
    with qtbot.waitSignal(header.sectionClicked):
        qtbot.mouseClick(header.viewport(), Qt.LeftButton, pos=QPoint(1, 1))

    # Assert we loaded all data before performing the sort.
    assert data(model, 0, 0) == 'i'


def test_filtering_with_large_rows(namespacebrowser, qtbot):
    """
    Test that filtering works when there's a large number of rows.
    """
    browser = namespacebrowser

    # Create data
    variables = {}
    for i in range(200):
        letter = string.ascii_lowercase[i // 10]
        var = letter + str(i)
        variables[var] = (
            {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
             'numpy_type': 'Unknown'}
        )

    # Set data
    browser.set_data(variables)

    # Assert we loaded the expected amount of data and that we can fetch
    # more data.
    model = browser.editor.model
    assert model.rowCount() == ROWS_TO_LOAD
    assert model.canFetchMore(QModelIndex())
    assert data(model, 49, 0) == 'e49'

    # Assert we can filter variables not loaded yet.
    browser.do_find("t19")
    assert model.rowCount() == 10

    # Assert all variables effectively start with 't19'.
    for i in range(10):
        assert data(model, i, 0) == 't19{}'.format(i)

    # Reset text_finder widget.
    browser.do_find('')

    # Create a new variable that starts with a different letter than
    # the rest.
    new_variables = variables.copy()
    new_variables['z'] = (
        {'type': 'int', 'size': 1, 'view': '1', 'python_type': 'int',
         'numpy_type': 'Unknown'}
    )

    # Emulate the process of loading those variables after the
    # namespace view is sent from the kernel.
    browser.process_remote_view(new_variables)

    # Assert that can find 'z' among the declared variables.
    browser.do_find("z")
    assert model.rowCount() == 1


def test_namespacebrowser_plot_with_mute_inline_plotting_true(
        namespacebrowser, qtbot):
    """
    Test that plotting a list from the namespace browser sends a signal
    with the plot if `mute_inline_plotting` is set to `True`.
    """
    namespacebrowser.set_conf('mute_inline_plotting', True, section='plots')
    namespacebrowser.plots_plugin_enabled = True
    my_list = [4, 2]
    mock_figure = Mock()
    mock_axis = Mock()
    mock_png = b'fake png'

    with patch('spyder.pyplot.subplots',
               return_value=(mock_figure, mock_axis)), \
         patch('IPython.core.pylabtools.print_figure',
               return_value=mock_png) as mock_print_figure, \
         qtbot.waitSignal(namespacebrowser.sig_show_figure_requested) \
             as blocker:
        namespacebrowser.plot(my_list, 'plot')

    mock_axis.plot.assert_called_once_with(my_list)
    mock_print_figure.assert_called_once_with(
        mock_figure, fmt='png', bbox_inches='tight', dpi=144)
    expected_args = [mock_png, 'image/png', namespacebrowser.shellwidget]
    assert blocker.args == expected_args


def test_namespacebrowser_plot_options(namespacebrowser):
    """
    Test that font.size and figure.subplot.bottom in matplotlib.rcParams are
    set to the values from the Spyder preferences when plotting.
    """
    def check_rc(*args):
        from matplotlib import rcParams
        assert rcParams['font.size'] == 20.5
        assert rcParams['figure.subplot.bottom'] == 0.314

    namespacebrowser.set_conf('mute_inline_plotting', True, section='plots')
    namespacebrowser.plots_plugin_enabled = True
    namespacebrowser.set_conf(
        'pylab/inline/fontsize', 20.5, section='ipython_console'
    )
    namespacebrowser.set_conf(
        'pylab/inline/bottom', 0.314, section='ipython_console'
    )

    mock_figure = Mock()
    mock_axis = Mock()
    mock_png = b'fake png'

    with patch('spyder.pyplot.subplots',
               return_value=(mock_figure, mock_axis)), \
         patch('IPython.core.pylabtools.print_figure',
               return_value=mock_png), \
         patch.object(mock_axis, 'plot', check_rc):
        namespacebrowser.plot([4, 2], 'plot')


def test_namespacebrowser_plot_with_mute_inline_plotting_false(namespacebrowser):
    """
    Test that plotting a list from the namespace browser shows a plot if
    `mute_inline_plotting` is set to `False`.
    """
    namespacebrowser.set_conf('mute_inline_plotting', False, section='plots')
    my_list = [4, 2]

    with patch('spyder.pyplot.figure') as mock_figure, \
         patch('spyder.pyplot.plot') as mock_plot, \
         patch('spyder.pyplot.show') as mock_show:
        namespacebrowser.plot(my_list, 'plot')

    mock_figure.assert_called_once_with()
    mock_plot.assert_called_once_with(my_list)
    mock_show.assert_called_once_with()


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""
Tests for the Help widgets.
"""

# Test library imports
import pytest

# Local imports
from spyder.plugins.help.widgets import RichText, PlainText


@pytest.fixture
def richtext(qtbot):
    """Set up richtext widget."""
    widget = RichText(None)
    qtbot.addWidget(widget)
    return widget


@pytest.fixture
def plaintext(qtbot):
    """Set up plaintext widget."""
    widget = PlainText(None)
    qtbot.addWidget(widget)
    return widget


def test_richtext(richtext):
    """Run RichText."""
    assert richtext


def test_plaintext(plaintext):
    """Run PlainText."""
    assert plaintext


if __name__ == "__main__":
    pytest.main()



# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Fixtures for the Spyder Remote Client plugin tests."""

from __future__ import annotations
from concurrent.futures import Future
import gc
import os
from pathlib import Path
import socket
import sys
import typing
import uuid

import pytest
from pytest_docker.plugin import Services
from qtpy.QtWidgets import QMainWindow

from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.app.cli_options import get_options
from spyder.config.manager import CONF
from spyder.plugins.ipythonconsole.plugin import IPythonConsole
from spyder.plugins.remoteclient.plugin import RemoteClient
from spyder.plugins.remoteclient.widgets import AuthenticationMethod


T = typing.TypeVar("T")

# NOTE: These credentials are hardcoded in the Dockerfile.
USERNAME = "ubuntu"
PASSWORD = USERNAME

HERE = Path(__file__).resolve().parent


def check_server(ip="127.0.0.1", port=22):
    """Check if a server is listening on the given IP and port.

    Args
    ----
        ip (str, optional): IP address to check. Defaults to "127.0.0.1".
        port (int, optional): Port to check. Defaults to 22.

    Returns
    -------
        bool: server is listening on the given IP and port
    """
    test_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        test_socket.connect((ip, port))
    except OSError:
        return False
    else:
        test_socket.close()
    return True


def await_future(future: Future[T], timeout: float = 30) -> T:
    """Wait for a future to finish or timeout."""
    return future.result(timeout=timeout)


class MainWindowMock(QMainWindow):
    """QMainWindow mock for the Remote Client plugin tests."""

    def __init__(self):
        # This avoids using the cli options passed to pytest
        sys_argv = [sys.argv[0]]
        self._cli_options = get_options(sys_argv)[0]
        super().__init__()
        PLUGIN_REGISTRY.set_main(self)

    def register_plugin(self, plugin_class):
        plugin = PLUGIN_REGISTRY.register_plugin(self, plugin_class)
        plugin._register()
        return plugin

    @staticmethod
    def unregister_plugin(plugin):
        assert PLUGIN_REGISTRY.delete_plugin(
            plugin.NAME
        ), f"{plugin.NAME} not deleted"
        plugin._unregister()

    @staticmethod
    def get_plugin(plugin_name, error=False):
        return PLUGIN_REGISTRY.get_plugin(plugin_name)

    @staticmethod
    def is_plugin_available(plugin_name):
        return PLUGIN_REGISTRY.is_plugin_available(plugin_name)


@pytest.fixture(scope="session")
def docker_compose_file(pytestconfig):
    """Override the default docker-compose.yml file."""
    return str(HERE / "docker-compose.yml")


@pytest.fixture()
def shell(ipyconsole, remote_client, remote_client_id, qtbot):
    """Create a new shell widget."""
    remote_client.create_ipyclient_for_server(remote_client_id)
    client = ipyconsole.get_current_client()
    shell = client.shellwidget
    qtbot.waitUntil(
        lambda: shell.spyder_kernel_ready and shell._prompt_html is not None,
        timeout=180000,
    )

    yield shell

    ipyconsole.get_widget().close_client(client=client)


@pytest.fixture(scope="class")
def remote_client_id(
    ssh_server_addr: typing.Tuple[str, int], remote_client: RemoteClient
) -> typing.Iterator[str]:
    """Add the Spyder Remote Client plugin to the registry."""
    ssh_options = {
        "host": ssh_server_addr[0],
        "port": ssh_server_addr[1],
        "username": USERNAME,
        "password": PASSWORD,
        "platform": "linux",
        "known_hosts": None,
    }
    config_id = str(uuid.uuid4())

    # Options Required by container widget
    remote_client.set_conf(
        f"{config_id}/auth_method", AuthenticationMethod.Password
    )
    remote_client.set_conf(
        f"{config_id}/{AuthenticationMethod.Password}/name", "test-server"
    )
    remote_client.set_conf(
        f"{config_id}/{AuthenticationMethod.Password}/address",
        ssh_server_addr[0],
    )
    remote_client.set_conf(
        f"{config_id}/{AuthenticationMethod.Password}/username", USERNAME
    )

    try:
        remote_client.load_client(options=ssh_options, config_id=config_id)
        yield config_id
    finally:
        remote_client.on_close(cancellable=False)


@pytest.fixture(scope="session")
def remote_client(
    ipyconsole_and_remoteclient: typing.Tuple[IPythonConsole, RemoteClient],
) -> RemoteClient:
    """
    Start the Spyder Remote Client plugin.

    Yields
    ------
        RemoteClient: Spyder Remote Client plugin.
    """
    return ipyconsole_and_remoteclient[1]


@pytest.fixture(scope="session")
def ipyconsole(
    ipyconsole_and_remoteclient: typing.Tuple[IPythonConsole, RemoteClient],
) -> IPythonConsole:
    """
    Start the IPython Console plugin.

    Yields
    ------
        IPythonConsole: IPython Console plugin.
    """
    return ipyconsole_and_remoteclient[0]


@pytest.fixture(scope="session")
def ipyconsole_and_remoteclient(qapp) -> (
    typing.Iterator[typing.Tuple[IPythonConsole, RemoteClient]]
):
    """
    Start the Spyder Remote Client plugin with IPython Console.

    Yields
    ------
        tuple: IPython Console and Spyder Remote Client plugins
    """

    window = MainWindowMock()

    os.environ["IPYCONSOLE_TESTING"] = "True"
    try:
        console = window.register_plugin(IPythonConsole)
        remote_client = window.register_plugin(RemoteClient)

        yield console, remote_client

        window.unregister_plugin(console)
        window.unregister_plugin(remote_client)
        del console, remote_client
    finally:
        os.environ.pop("IPYCONSOLE_TESTING")
        CONF.reset_manager()
        PLUGIN_REGISTRY.reset()
        del window
        gc.collect()


@pytest.fixture(scope="class")
def ssh_server_addr(
    docker_ip: str, docker_services: Services
) -> typing.Tuple[str, int]:
    """Start an SSH server from docker-compose and return its address.

    Args
    ----
        docker_ip (str): IP address of the Docker daemon.
        docker_services (Services): Docker services.

    Returns
    -------
        tuple: IP address and port of the SSH server.
    """
    # Build URL to service listening on random port.
    # NOTE: This is the service name and port in the docker-compose.yml file.
    port = docker_services.port_for("test-spyder-remote-server", 22)

    docker_services.wait_until_responsive(
        check=lambda: check_server(ip=docker_ip, port=port),
        timeout=30.0,
        pause=0.1,
    )

    return docker_ip, port



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for SidebarDialog."""

# Third party imports
from qtpy.QtWidgets import QLabel, QVBoxLayout
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.utils.stylesheet import APP_STYLESHEET
from spyder.widgets.sidebardialog import SidebarDialog, SidebarPage


# --- Fixtures
# -----------------------------------------------------------------------------
@pytest.fixture
def sidebar_dialog(qapp, qtbot):

    # Pages
    class Page1(SidebarPage):

        def get_name(self):
            return "Page 1"

        def get_icon(self):
            return self.create_icon("variable_explorer")

        def setup_page(self):
            self.label = QLabel("This is page one!")
            layout = QVBoxLayout()
            layout.addWidget(self.label)
            layout.addStretch(1)
            self.setLayout(layout)

    class Page2(SidebarPage):

        def get_name(self):
            return "Page 2"

        def get_icon(self):
            return self.create_icon("files")

        def setup_page(self):
            self.label = QLabel("This is page two!")
            layout = QVBoxLayout()
            layout.addWidget(self.label)
            layout.addStretch(1)
            self.setLayout(layout)

    # Dialog
    class TestDialog(SidebarDialog):
        PAGE_CLASSES = [Page1, Page2]

    if not running_in_ci():
        qapp.setStyleSheet(str(APP_STYLESHEET))
    dialog = TestDialog()
    qtbot.addWidget(dialog)

    # To check the dialog visually
    with qtbot.waitExposed(dialog):
        dialog.show()

    return dialog


# --- Tests
# -----------------------------------------------------------------------------
def test_sidebardialog(sidebar_dialog, qtbot):
    dialog = sidebar_dialog
    assert dialog is not None

    # Check label displayed in the initial page
    assert "one" in dialog.get_page().label.text()

    # Check label in the second page
    dialog.set_current_index(1)
    assert "two" in dialog.get_page().label.text()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Tests for the `spyder.config`."""

# Standard library imports
import os

# Third party imports
import pytest

# Local imports
from spyder.config.base import get_conf_path
from spyder.config.user import DefaultsConfig, SpyderUserConfig, UserConfig


@pytest.fixture
def defaultconfig(tmpdir, request):
    name = 'defaults-test'
    path = str(tmpdir)
    default_kwargs = {'name': name, 'path': path}

    param = getattr(request, 'param', None)
    if param:
        modified_kwargs = request.param[0]
        kwargs = default_kwargs.copy().update(modified_kwargs)
    else:
        kwargs = default_kwargs

    return DefaultsConfig(**kwargs)


@pytest.fixture
def userconfig(tmpdir, request):
    ini_contents = '[main]\nversion = 1.0.0\n\n'
    ini_contents += "[section]\noption = value\n\n"

    name = 'spyder-test'
    path = str(tmpdir)
    default_kwargs = {
        'name': name,
        'path': path,
        'defaults': {},
        'load': True,
        'version': '1.0.0',
        'backup': False,
        'raw_mode': True,
        'remove_obsolete': False,
    }

    param = getattr(request, 'param', None)
    if param:
        modified_kwargs = request.param[0]
        kwargs = default_kwargs.copy().update(modified_kwargs)
    else:
        kwargs = default_kwargs

    inifile = tmpdir.join('{}.ini'.format(name))
    inifile.write(ini_contents)

    return UserConfig(**kwargs)


@pytest.fixture
def spyderconfig(tmpdir, request):
    ini_contents = '[main]\nversion = 1.0.0\n\n'
    ini_contents += """[ipython_console]
startup/run_lines = value1,value2

"""

    name = 'spyder-test'
    path = str(tmpdir)
    default_kwargs = {
        'name': name,
        'path': path,
        'defaults': {},
        'load': True,
        'version': '1.0.0',
        'backup': False,
        'raw_mode': True,
        'remove_obsolete': False,
    }

    param = getattr(request, 'param', None)
    if param:
        modified_kwargs = param[0]
        kwargs = default_kwargs.copy()
        kwargs.update(modified_kwargs)
    else:
        kwargs = default_kwargs

    inifile = tmpdir.join('{}.ini'.format(name))
    inifile.write(ini_contents)

    return SpyderUserConfig(**kwargs)


@pytest.fixture
def spyderconfig_patches_42(tmpdir):
    ini_contents = '[main]\nversion = 42.0.0\n\n'
    ini_contents += '[ipython_console]\nstartup/run_lines = value1,value2'

    name = 'spyder'
    inifile = tmpdir.join('{}.ini'.format(name))
    inifile.write(ini_contents + '\n\n')

    return SpyderUserConfig(name=name, path=str(tmpdir), defaults={},
                            load=True, version='43.0.0', backup=False,
                            raw_mode=True, remove_obsolete=False)


@pytest.fixture
def spyderconfig_patches_45(tmpdir):
    ini_contents = '[main]\nversion = 45.0.0\n\n'
    ini_contents += '[ipython_console]\nstartup/run_lines = value1,value2'

    name = 'spyder'
    inifile = tmpdir.join('{}.ini'.format(name))
    inifile.write(ini_contents + '\n\n')

    return SpyderUserConfig(name=name, path=str(tmpdir), defaults={},
                            load=True, version='46.0.0', backup=False,
                            raw_mode=True, remove_obsolete=False)


@pytest.fixture
def spyderconfig_previous(tmpdir, mocker):
    ini_contents = '[main]\nversion = 50.0.0\n\n'
    ini_contents += "[section]\noption = value"

    name = 'spyder'
    path = str(tmpdir)

    def temp(*args, **kwargs):
        return path

    mocker.patch('spyder.config.base', 'get_conf_path')
    inifile = tmpdir.join('{}.ini'.format(name))
    os.makedirs(str(tmpdir.join('app', 'config')))
    inifile2 = tmpdir.join('app', 'config', '{}.ini'.format(name))
    inifile.write(ini_contents + '\n\n')
    inifile2.write(ini_contents + '\n\n')

    return SpyderUserConfig(name=name, path=path, defaults={}, load=True,
                            version='50.0.0', backup=False, raw_mode=True,
                            remove_obsolete=False)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""
Tests for module_completion.py
"""

# Stdlib imports
import sys

# Test library imports
import pytest

# Local imports
from spyder.utils.introspection.module_completion import get_preferred_submodules


@pytest.mark.skipif(sys.platform == 'darwin',
                    reason="It's very slow on Mac")
def test_module_completion():
    """Test module_completion."""
    assert 'numpy.linalg' in get_preferred_submodules()


if __name__ == "__main__":
    pytest.main()
